from typing import List

import bw2data as bd
import bw_functional as bf

from activity_browser import bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class ExchangeSDFToClipboard(ABAction):
    """
    ABAction to open one or more supplied activities in an activity tab by employing signals.

    TODO: move away from using signals like this. Probably add a method to the MainWindow to add a panel instead.
    """

    icon = qicons.superstructure
    text = "SDF to clipboard"

    @staticmethod
    @exception_dialogs
    def run(exchanges: List[int | bd.Edge]):
        exchanges = [bwutils.refresh_edge(edge) for edge in exchanges]

        virtual_exchanges = []
        for exchange in exchanges:
            if isinstance(exchange, bf.MFExchange):
                virtual_exchanges += exchange.virtual_edges
            else:
                virtual_exchanges.append(exchange.as_dict())

        df = bwutils.exchanges_to_sdf(virtual_exchanges)
        df.to_clipboard(excel=True, index=False)



from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from activity_browser.ui.wizards.settings_wizard import SettingsWizard


class SettingsWizardOpen(ABAction):
    """ABAction to open the SettingsWizard"""

    icon = qicons.settings
    text = "Settings..."

    @staticmethod
    @exception_dialogs
    def run():
        SettingsWizard(application.main_window).show()



from .database_explorer import DatabaseExplorerPane
from .database_products import DatabaseProductsPane
from .project_manager import ProjectManagerPane
from .databases import DatabasesPane
from .impact_categories import ImpactCategoriesPane
from .calculation_setups import CalculationSetupsPane


registered_panes = [
    DatabaseExplorerPane,
    DatabaseProductsPane,
    ProjectManagerPane,
    DatabasesPane,
    ImpactCategoriesPane,
    CalculationSetupsPane,
]

shown_panes = [
    DatabasesPane,
    ImpactCategoriesPane,
    CalculationSetupsPane,
]

hidden_panes = [
    ProjectManagerPane,
]

default_panes = shown_panes + hidden_panes



import functools

blacklist = ["__module__", "__subclasshook__", "__dict__", "__init_subclass__"]


def patch_superclass(cls):
    """
    Class decorator that will patch any methods that differ from the superclass onto said superclass during runtime.
    """
    # collect the superclass
    superclass = cls.__bases__[0]

    # iterate over the attributes in the decorated class
    for name, value in [(name, getattr(cls, name)) for name in dir(cls)]:
        try:
            # return if the name is in the blacklist
            if name in blacklist:
                continue

            # return if the attribute is the same as the one in the superclass
            if value == getattr(superclass, name, None):
                continue

            # saving the attribute that we'll be patching
            patched[superclass] = patched.get(superclass, {})
            patched[superclass].update({name: getattr(superclass, name, None)})

            # patching the attribute
            setattr(superclass, name, value)
        except Exception as e:
            # an exception will be interesting, but patching shouldn't break, so continue.
            print(e)
            pass

    return cls


def patch_attribute(obj, name):
    """
    Product decorator to patch single attributes of a class. Handy when the superclass is already patched by
    another library and will be too different to use patch_superclass on. Pass the class you want to patch as argument
    """

    def inner(obj, name, fn):
        # saving the attribute that we'll be patching
        patched[obj] = patched.get(obj, {})
        patched[obj].update({name: getattr(obj, name, None)})

        setattr(obj, name, fn)
        return

    return functools.partial(inner, obj, name)


class Patched(dict):
    def __getitem__(self, obj):
        """
        We subclass this method because we also want subclasses to work for this.
        """
        if obj in self:
            return super().__getitem__(obj)
        elif match := [
            key for key in self if issubclass(obj, key) or isinstance(obj, key)
        ]:
            return super().__getitem__(match[0])
        raise KeyError("Patched object not found")


# dictionary of patched attributes. Access a patched attribute like so: patched[class][attribute_name]
patched = Patched()



from logging import getLogger

from qtpy import QtCore, QtWidgets

import bw2data as bd

from activity_browser import signals, bwutils
from activity_browser.ui import widgets

from .activity_header import ActivityHeader
from .exchanges_tab import ExchangesTab
from .description_tab import DescriptionTab
from .graph_tab import GraphTab
from .parameters_tab import ParametersTab
from .data_tab import DataTab
from .consumers_tab import ConsumersTab

log = getLogger(__name__)


class ActivityDetailsPage(QtWidgets.QWidget):
    """
    A widget that displays detailed information about a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display details for.
        activity_data_grid (ActivityHeader): The header widget displaying activity data.
        tabs (QtWidgets.QTabWidget): The tab widget containing various detail tabs.
        exchanges_tab (ExchangesTab): The tab displaying exchanges related to the activity.
        description_tab (DescriptionTab): The tab displaying the description of the activity.
        graph_explorer (GraphTab): The tab displaying the graph related to the activity.
        parameters_tab (ParametersTab): The tab displaying parameters of the activity.
        consumer_tab (ConsumersTab): The tab displaying consumers of the activity.
        data_tab (DataTab): The tab displaying data related to the activity.
    """
    _populate_later_flag = False

    def __init__(self, activity: tuple | int | bd.Node, parent=None):
        """
        Initializes the ActivityDetailsPage widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display details for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self.activity = bd.get_activity(activity)
        self.setObjectName(f"activity_details_{self.activity['database']}_{self.activity['code']}")
        self.setWindowTitle(self.activity["name"])

        # Initialize header widget for activity data
        self.activity_data_grid = ActivityHeader(self)
        # Initialize tab widget to hold various detail tabs
        self.tabs = QtWidgets.QTabWidget(self)

        # Initialize and add the Exchanges tab
        self.exchanges_tab = ExchangesTab(activity, self)
        self.tabs.addTab(self.exchanges_tab, "Exchanges")

        # Initialize and add the Description tab
        self.description_tab = DescriptionTab(activity, self)
        self.tabs.addTab(self.description_tab, "Description")

        # Initialize and add the Graph tab
        self.graph_explorer = GraphTab(activity, self)
        self.tabs.addTab(self.graph_explorer, "Graph")

        # Initialize and add the Parameters tab
        self.parameters_tab = ParametersTab(activity, self)
        self.tabs.addTab(self.parameters_tab, "Parameters")

        # Initialize and add the Consumers tab
        self.consumer_tab = ConsumersTab(activity, self)
        self.tabs.addTab(self.consumer_tab, "Consumers")

        # Initialize and add the Data tab
        self.data_tab = DataTab(activity, self)
        self.tabs.addTab(self.data_tab, "Data")

        # Build the layout of the widget
        self.build_layout()
        # Synchronize the widget with the current state of the activity
        self.sync()
        # Connect signals to their respective slots
        self.connect_signals()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(10, 10, 4, 1)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        # Add the activity data grid and tabs to the layout
        layout.addWidget(self.activity_data_grid)
        layout.addWidget(widgets.ABHLine(self))
        layout.addWidget(self.tabs)

        self.setLayout(layout)

    def connect_signals(self):
        """
        Connects signals to their respective slots.
        """
        signals.node.deleted.connect(self.on_node_deleted)
        signals.database.deleted.connect(self.on_database_deleted)
        signals.meta.databases_changed.connect(self.syncLater)
        signals.parameter.recalculated.connect(self.syncLater)
        signals.node.changed.connect(self.syncLater)

    def on_node_deleted(self, node):
        """
        Slot to handle node deletion.

        Args:
            node: The node that was deleted.
        """
        if node.id == self.activity.id:
            self.deleteLater()

    def on_database_deleted(self, name):
        """
        Slot to handle database deletion.

        Args:
            name: The name of the database that was deleted.
        """
        if name == self.activity["database"]:
            self.deleteLater()

    def syncLater(self):
        """
        Schedules a sync operation to be performed later.
        """

        def slot():
            self._populate_later_flag = False
            self.sync()
            self.thread().eventDispatcher().awake.disconnect(slot)

        if self._populate_later_flag:
            return

        self._populate_later_flag = True
        self.thread().eventDispatcher().awake.connect(slot)

    def sync(self):
        """
        Synchronizes the widget with the current state of the activity.
        """
        self.activity = bwutils.refresh_node_or_none(self.activity)

        if self.activity is None:
            # Activity was already deleted
            return

        # Update the tab name to be the activity name
        self.setWindowTitle(self.activity["name"])

        # Synchronize all tabs with the current state of the activity
        self.activity_data_grid.sync()
        self.exchanges_tab.sync()
        self.description_tab.sync()
        self.consumer_tab.sync()
        self.data_tab.sync()
        self.parameters_tab.sync()



from qtpy import QtCore, QtWidgets

import bw2data as bd
from bw2data.parameters import Group
from bw2data.backends.proxies import ExchangeDataset, Exchanges

from activity_browser import application, settings
from activity_browser.bwutils import AB_metadata
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class DatabaseDelete(ABAction):
    """
    Deletes a specified database from the project after user confirmation.

    This method performs the following steps:
    - Displays a confirmation dialog to the user with the database name and record count.
    - If the user confirms, deletes the database, its upstream exchanges, and associated parameters.
    - Removes the database from the project settings.

    Args:
        db_name (str): The name of the database to be deleted.

    Steps:
    - Set the cursor to a waiting state while gathering data for large databases.
    - Retrieve the record count for the specified database.
    - Construct a warning message with the database name and record count.
    - Display a confirmation dialog to the user.
    - If the user cancels, exit the method.
    - Set the cursor to a waiting state while performing the deletion.
    - Delete upstream exchanges associated with the database.
    - Remove the database from the Brightway2 project.
    - Delete database parameters.
    - Remove the database from the project settings.
    - Restore the cursor to its default state.
    """

    icon = qicons.delete
    text = "Delete database"
    tool_tip = "Delete this database from the project"

    @staticmethod
    @exception_dialogs
    def run(db_name: str):
        # gathering data will take time for large databases
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        # get the record count from the database controller
        db_name = db_name
        n_records = AB_metadata.dataframe[AB_metadata.dataframe["database"] == db_name].shape[0]

        # construct warning text
        text = f"Are you sure you want to delete database '{db_name}'?"
        if n_records:
            text += f" It contains {n_records} activities"

        # ask the user for confirmation
        QtWidgets.QApplication.restoreOverrideCursor()
        response = QtWidgets.QMessageBox.question(
            application.main_window, "Delete database?", text
        )

        # return if the user cancels
        if response != response.Yes:
            return

        # deleting data will take time for large databases
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        # delete upstream exchanges
        ExchangeDataset.delete().where(ExchangeDataset.input_database == db_name).execute()

        # instruct the DatabaseController to delete the database from the project.
        del bd.databases[db_name]

        # delete database parameters
        Group.delete().where(Group.name == db_name).execute()

        # remove database from project settings
        settings.project_settings.remove_db(db_name)

        QtWidgets.QApplication.restoreOverrideCursor()



import numbers
import json
from datetime import datetime as dt
from pathlib import Path
from typing import Union

import xlsxwriter
from bw2io.export.csv import reformat
from bw2io.export.excel import CSVFormatter, create_valid_worksheet_name

from activity_browser.mod import bw2data as bd

from .importers import ABPackage
from .pedigree import PedigreeMatrix

# Copied most of this code wholesale from bw2io package.
# TODO: reminder to make a pull-request for these things in bw2io repo.
#  - Add the 'nan_inf_to_errors' option when opening the xlsxwriter.Workbook.
#  - Add handler for pedigree data to exporter
#  - Add 'database' field as required CSVFormatter export field.
#  - Add code to ensure no second 'activity' field is exported, as this
#  messes with following import.


def ab_reformat(value):
    if isinstance(value, dict):
        try:
            return json.dumps(value)
        except TypeError:
            return "Non-serializable dictionary"
    return reformat(value)


class ABCSVFormatter(CSVFormatter):
    def get_activity_metadata(self, act):
        excluded = {"database", "name", "activity"}
        return {
            "name": act.get("name"),
            "metadata": sorted(
                [
                    (k, ab_reformat(v))
                    for k, v in act.items()
                    if k not in excluded and not isinstance(v, list)
                ]
            ),
            "parameters": self.get_activity_parameters(act),
        }

    def exchange_as_dict(self, exc):
        """Same as CSVFormatter, but explicitly pull the database from the
        input activity.

        This ensures that the database value is always included, even when
        it is not stored in the exchange _data.
        """
        inp = exc.input
        inp_fields = (
            "name",
            "unit",
            "location",
            "categories",
            "database",
            "reference product",
        )
        skip_fields = ("input", "output")
        data = {k: v for k, v in exc._data.items() if k not in skip_fields}
        if "product" in data and "reference product" not in data:
            data["reference product"] = data.pop("product")
        data.update(**{k: inp[k] for k in inp_fields if inp.get(k)})
        return data


def format_pedigree(data: dict) -> str:
    """Converts pedigree dict to tuple."""
    try:
        return "::".join(
            str(x) for x in PedigreeMatrix.from_dict(data).factors_as_tuple()
        )
    except AssertionError:
        # Will skip any other kinds of dictionaries.
        return "nan"


def frmt_str(data: Union[str, dict]) -> str:
    """Format non-numerical data (like tuples) to string format."""
    return format_pedigree(data) if isinstance(data, dict) else str(data)


def write_lci_excel(db_name: str, path: str, objs=None, sections=None) -> Path:
    """Export database `database_name` to an Excel spreadsheet.

    Not all data can be exported. The following constraints apply:

    * Nested data, e.g. `{'foo': {'bar': 'baz'}}` are excluded. Spreadsheets are not a great format for nested data.
      However, *tuples* are exported, and the characters `::` are used to join elements of the tuple.
    * The only well-supported data types are strings, numbers, and booleans.

    Returns the filepath of the exported file.

    """
    path = Path(path)
    if not path.suffix == ".xlsx":
        out_file = path / "lci-{}.xlsx".format(bd.utils.safe_filename(db_name, False))
    else:
        out_file = path

    workbook = xlsxwriter.Workbook(out_file, {"nan_inf_to_errors": True})
    bold = workbook.add_format({"bold": True})
    bold.set_font_size(12)
    highlighted = {
        "Activity",
        "Database",
        "Exchanges",
        "Parameters",
        "Database parameters",
        "Project parameters",
    }
    frmt = lambda x: bold if row[0] in highlighted else None

    sheet = workbook.add_worksheet(create_valid_worksheet_name(db_name))

    data = ABCSVFormatter(db_name, objs).get_formatted_data(sections)

    for row_index, row in enumerate(data):
        for col_index, value in enumerate(row):
            if value is None:
                continue
            elif isinstance(value, numbers.Number):
                sheet.write_number(row_index, col_index, value, frmt(value))
            else:
                sheet.write_string(row_index, col_index, frmt_str(value), frmt(value))

    workbook.close()

    return out_file


def store_database_as_package(db_name: str, directory: str = None) -> bool:
    """Attempt to use `bw.BW2Package` to save the given database as an
    isolated package that can be shared with others.
    Returns a boolean signifying success or failure.
    """
    if db_name not in bd.databases:
        return False
    metadata = bd.databases[db_name]
    db = bd.Database(db_name)
    directory = directory or bd.projects.output_dir
    output_dir = Path(directory)
    if output_dir.suffix == ".bw2package":
        out_file = output_dir
    else:
        out_file = output_dir / "{}.bw2package".format(db.filename)
    # First, ensure the metadata on the database is up-to-date.
    modified = dt.strptime(metadata["modified"], "%Y-%m-%dT%H:%M:%S.%f")
    if "processed" in metadata:
        processed = dt.strptime(metadata["processed"], "%Y-%m-%dT%H:%M:%S.%f")
        if processed < modified:
            db.process()
    else:
        db.process()
    # Now that processing is done, perform the export.
    ABPackage.unrestricted_export(db, out_file)
    return True



# -*- coding: utf-8 -*-
import json
import os
import shutil
from pathlib import Path
from typing import Optional, Any
from logging import getLogger

import bw2data as bd

import platformdirs
from qtpy.QtWidgets import QMessageBox

from .signals import signals

log = getLogger(__name__)
DEFAULT_BW_DATA_DIR = bd.projects._base_data_dir


def pathlib_encoder(value: Any) -> Any:
    if isinstance(value, Path):
        return str(value)
    else:
        return value


class BaseSettings(object):
    """Base Class for handling JSON settings files."""

    def __init__(self, directory: str, filename: str = None):
        self.data_dir = directory
        self.filename = filename or "default_settings.json"
        self.settings_file = os.path.join(self.data_dir, self.filename)
        self.settings: Optional[dict] = None
        self.initialize_settings()

    @classmethod
    def get_default_settings(cls) -> dict:
        """Returns dictionary containing the default settings for the file"""
        raise NotImplementedError

    def restore_default_settings(self) -> None:
        """Undo all user settings and return to original state."""
        self.settings = self.get_default_settings()
        self.write_settings()

    def initialize_settings(self) -> None:
        """Attempt to find and read the settings_file, creates a default
        if not found
        """
        if os.path.isfile(self.settings_file):
            self.load_settings()
        else:
            self.settings = self.get_default_settings()
            self.write_settings()

    def load_settings(self) -> None:
        with open(self.settings_file, "r") as infile:
            self.settings = json.load(infile)

    def write_settings(self) -> None:
        with open(self.settings_file, "w") as outfile:
            json.dump(self.settings, outfile, indent=4, sort_keys=True, default=pathlib_encoder)


class ABSettings(BaseSettings):
    """
    Interface to the json settings file. Will create a userdata directory via platformdirs if not
    already present.
    """

    def __init__(self, filename: str):
        ab_dir = str(platformdirs.user_data_dir(appname="ActivityBrowser", appauthor="ActivityBrowser"))
        if not os.path.isdir(ab_dir):
            os.makedirs(ab_dir, exist_ok=True)
        self.update_old_settings(ab_dir, filename)

        # Currently loaded plugins objects as:
        # {plugin_name: <plugin_object>, ...}
        # this list is generated at startup and never writen in settings.
        # it is filled by the plugin controller
        self.plugins = {}

        super().__init__(ab_dir, filename)

        if not self.healthy():
            log.warn("Settings health check failed, resetting")
            self.restore_default_settings()

    def healthy(self) -> bool:
        """
        Checks the settings file to see if it is healthy. Returns True if all checks pass, otherwise returns False.
        """
        healthy = True

        # check for write access to the current bw dir
        healthy = healthy and os.access(self.settings.get("current_bw_dir"), os.W_OK)

        # check for write access to the custom bw dirs
        access = [os.access(path, os.W_OK) for path in self.settings.get("custom_bw_dirs")]
        healthy = healthy and False not in access

        return healthy

    @staticmethod
    def update_old_settings(directory: str, filename: str) -> None:
        """Recycling code to enable backward compatibility: This function is only required for compatibility
        with the old settings file and can be removed in a future release
        """
        file = os.path.join(directory, filename)
        if not os.path.exists(file):
            package_dir = Path(__file__).resolve().parents[1]
            old_settings = os.path.join(package_dir, "ABsettings.json")
            if os.path.exists(old_settings):
                shutil.copyfile(old_settings, file)
        if os.path.isfile(file):
            with open(file, "r") as current:
                current_settings = json.load(current)
            if "current_bw_dir" not in current_settings:
                new_settings_content = {
                    "current_bw_dir": current_settings["custom_bw_dir"],
                    "custom_bw_dirs": [current_settings["custom_bw_dir"]],
                    "startup_project": current_settings["startup_project"],
                }
                with open(file, "w") as new_file:
                    json.dump(new_settings_content, new_file, default=pathlib_encoder)

    @classmethod
    def get_default_settings(cls) -> dict:
        """Using methods from the commontasks file to set default settings"""
        return {
            "current_bw_dir": cls.get_default_directory(),
            "custom_bw_dirs": [cls.get_default_directory()],
            "startup_project": cls.get_default_project_name(),
        }

    @property
    def custom_bw_dir(self) -> str:
        """Returns the custom brightway directory, or the default"""
        return self.settings.get("custom_bw_dirs", self.get_default_directory())

    @property
    def current_bw_dir(self) -> str:
        """Returns the current brightway directory"""
        return self.settings.get("current_bw_dir", self.get_default_directory())

    @current_bw_dir.setter
    def current_bw_dir(self, directory: str) -> None:
        self.settings["current_bw_dir"] = directory
        self.write_settings()

    @custom_bw_dir.setter
    def custom_bw_dir(self, directory: str) -> None:
        """Sets the custom brightway directory to `directory`"""
        if directory not in self.settings["custom_bw_dirs"]:
            self.settings["custom_bw_dirs"].append(directory)
            self.write_settings()

    def remove_custom_bw_dir(self, directory: str) -> None:
        """Removes the brightway directory to 'directory'"""
        try:
            self.settings["custom_bw_dirs"].remove(directory)
            self.write_settings()
        except KeyError as e:
            QMessageBox.warning(
                self,
                f"Error while attempting to remove a brightway environmental dir: {e}",
            )

    @property
    def startup_project(self) -> str:
        """Get the startup project from the settings, or the default"""
        project = self.settings.get("startup_project", self.get_default_project_name())
        if project and project not in bd.projects:
            project = self.get_default_project_name()
        return project

    @startup_project.setter
    def startup_project(self, project: str) -> None:
        """Sets the startup project to `project`"""
        self.settings.update({"startup_project": project})

    @staticmethod
    def get_default_directory() -> str:
        """Returns the default brightway application directory"""
        return DEFAULT_BW_DATA_DIR

    @staticmethod
    def get_default_project_name() -> Optional[str]:
        """Returns the default project name."""
        if "default" in bd.projects:
            return "default"
        elif len(bd.projects):
            return next(iter(bd.projects)).name
        else:
            return None

    @property
    def theme(self) -> str:
        """Returns the current brightway directory"""
        return self.settings.get("theme", "Light theme")

    @theme.setter
    def theme(self, new_theme: str) -> None:
        self.settings.update({"theme": new_theme})


class ProjectSettings(BaseSettings):
    """
    Handles user settings which are specific to projects. Created initially to handle read-only/writable database status
    Code based on ABSettings class, if more different types of settings are needed, could inherit from a base class

    structure: singleton, loaded dependent on which project is selected.
        Persisted on disc, Stored in the BW2 projects data folder for each project
        a dictionary1 of dictionaries2
        Dictionary1 keys are settings names (currently just 'read-only-databases'), values are dictionary2s
        Dictionary2 keys are database names, values are bools

    For now, decided to not include saving writable-activities to settings.
    As activities are identified by tuples, and saving them to json requires extra code
    https://stackoverflow.com/questions/15721363/preserve-python-tuples-with-json
    This is currently not worth the effort but could be returned to later

    """

    def __init__(self, filename: str):
        # on selection of a project (signal?), find the settings file for that project if it exists
        # it can be a custom location, based on ABsettings. So check that, and if not, use default?
        # once found, load the settings or just an empty dict.
        self.connect_signals()
        super().__init__(bd.projects.dir, filename)

        bd.projects.dir.joinpath("activity_browser").mkdir(exist_ok=True)

        # https://github.com/LCA-ActivityBrowser/activity-browser/issues/235
        # Fix empty settings file and populate with currently active databases
        if "read-only-databases" not in self.settings:
            self.settings.update(self.process_brightway_databases())
            self.write_settings()
        if "plugins_list" not in self.settings:
            self.settings.update({"plugins_list": []})
            self.write_settings()

    def connect_signals(self):

        # Reload the project settings whenever a project switch occurs.
        signals.project.changed.connect(self.reset_for_project_selection)

        # save new plugin for this project
        signals.plugin_selected.connect(self.add_plugin)

    @classmethod
    def get_default_settings(cls) -> dict:
        """Return default empty settings dictionary."""
        settings = cls.process_brightway_databases()
        settings["plugins_list"] = []
        return settings

    @staticmethod
    def process_brightway_databases() -> dict:
        """Process brightway database list and return new settings dictionary.

        NOTE: This ignores the existing database read-only settings.
        """
        return {"read-only-databases": {name: True for name in bd.databases.list}}

    def reset_for_project_selection(self) -> None:
        """On switching project, attempt to read the settings for the new
        project.
        """
        log.info(f"Project settings directory: {bd.projects.dir}")

        bd.projects.dir.joinpath("activity_browser").mkdir(exist_ok=True)

        self.settings_file = os.path.join(bd.projects.dir, self.filename)
        self.initialize_settings()
        # create a plugins_list entry for old projects
        if "plugins_list" not in self.settings:
            self.settings.update({"plugins_list": []})
            self.write_settings()

    def add_db(self, db_name: str, read_only: bool = True) -> None:
        """Store new databases and relevant settings here when created/imported"""
        self.settings["read-only-databases"].setdefault(db_name, read_only)
        self.write_settings()

    def modify_db(self, db_name: str, read_only: bool) -> None:
        """Update write-rules for the given database"""
        self.settings["read-only-databases"].update({db_name: read_only})
        self.write_settings()

    def remove_db(self, db_name: str) -> None:
        """When a database is deleted from a project, the settings are also deleted."""
        self.settings["read-only-databases"].pop(db_name, None)
        self.write_settings()

    def db_is_readonly(self, db_name: str) -> bool:
        """Check if given database is read-only, defaults to yes."""
        return self.settings["read-only-databases"].get(db_name, True)

    def get_editable_databases(self):
        """Return list of database names where read-only is false

        NOTE: discards the biosphere3 database based on name.
        """
        iterator = self.settings.get("read-only-databases", {}).items()
        return (name for name, ro in iterator if not ro and name != "biosphere3")

    def add_plugin(self, name: str, select: bool = True):
        """Add a plugin to settings or remove it"""
        if select:
            self.settings["plugins_list"].append(name)
            self.write_settings()
            return
        if name in self.settings["plugins_list"]:
            self.settings["plugins_list"].remove(name)
            self.write_settings()

    def get_plugins_list(self):
        """Return a list of plugins names"""
        return self.settings["plugins_list"]


ab_settings = ABSettings("ABsettings.json")
project_settings = ProjectSettings("AB_project_settings.json")



from qtpy import QtCore, QtGui, QtWidgets


class ABListEditDialog(QtWidgets.QDialog):
    def __init__(self, data, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Edit List/Tuple")
        self.resize(400, 300)

        layout = QtWidgets.QVBoxLayout(self)

        # Table
        self.list = QtWidgets.QListWidget(self)
        self.list.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.list.setDefaultDropAction(QtGui.Qt.MoveAction)
        self.list.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.list.setEditTriggers(QtWidgets.QAbstractItemView.DoubleClicked | QtWidgets.QAbstractItemView.EditKeyPressed)

        layout.addWidget(self.list)

        # OK/Cancel
        self.button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        layout.addWidget(self.button_box)

        # Signals
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        self.load_data(data)

        self.list.itemChanged.connect(self.on_item_changed)

    def load_data(self, data):
        for value in data:
            self._append_item(str(value))
        self._append_add_field()

    def _append_item(self, text=""):
        item = QtWidgets.QListWidgetItem(text)
        item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)
        self.list.addItem(item)

    def _append_add_field(self):
        """Always keep a final row with placeholder text."""
        item = QtWidgets.QListWidgetItem("add field")
        item.setFlags(QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable)
        item.setForeground(QtCore.Qt.gray)
        self.list.addItem(item)

    def on_item_changed(self, item: QtWidgets.QListWidgetItem):
        row = self.list.row(item)
        last_row = self.list.count() - 1

        # User edited the placeholder row -> convert it to real row and add new placeholder
        if row == last_row and item.text().strip() and item.text() != "add field":
            self._append_add_field()
            item.setFlags(item.flags() | QtCore.Qt.ItemIsDragEnabled)
            item.setForeground(QtCore.Qt.black)
            return

        # If user clears a normal row, remove it
        elif row != last_row and not item.text().strip():
            self.list.takeItem(row)
            return

        # Restore placeholder if someone empties the last row
        elif row == last_row and not item.text().strip():
            item.setForeground(QtCore.Qt.gray)
            item.setText("add field")
            return

    def get_data(self, as_tuple=False):
        values = []
        last_row = self.list.count() - 1
        for row in range(self.list.count()):
            if row == last_row:
                continue  # skip the "add field" row
            item = self.list.item(row)
            if item and item.text().strip():
                values.append(item.text().strip())
        return tuple(values) if as_tuple else values





from qtpy import QtWidgets, QtGui

import bw2data as bd
import pandas as pd

from activity_browser import signals, actions
from activity_browser.ui import widgets, delegates


class CalculationSetupsPane(widgets.ABAbstractPane):
    title = "Calculation Setups"
    unique = True

    def __init__(self, parent):
        """
        Initializes the CalculationSetupsPane.

        This constructor sets up the view and model for displaying calculation setups,
        configures the view's appearance and behavior, and builds the layout while
        connecting necessary signals.

        Args:
            parent (QtWidgets.QWidget): The parent widget for this pane.
        """
        super().__init__(parent)
        self.view = CalculationSetupsView()
        self.model = CalculationSetupsModel()
        self.view.setModel(self.model)

        self.view.setAlternatingRowColors(True)

        self.build_layout()
        self.connect_signals()

    def connect_signals(self):
        """
        Connects the signals to the appropriate slots.
        """
        signals.meta.calculation_setups_changed.connect(self.sync)
        signals.project.changed.connect(self.sync)

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(5, 0, 5, 5)
        self.setLayout(layout)

    def sync(self):
        """
        Synchronizes the model with the current state of the calculation setups.
        """
        self.model.setDataFrame(self.build_df())
        self.view.resizeColumnToContents(0)

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from the calculation setups.

        Returns:
            pd.DataFrame: The DataFrame containing the calculation setups data.
        """
        data = []
        for cs in bd.calculation_setups:
            data.append(
                {
                    "name": cs,
                    "functional_units": len(bd.calculation_setups[cs].get("inv", [])),
                    "impact_categories": len(bd.calculation_setups[cs].get("ia", [])),
                }
            )

        cols = ["name", "functional_units", "impact_categories"]

        return pd.DataFrame(data, columns=cols)


class CalculationSetupsView(widgets.ABTreeView):
    """
    A view that displays the calculation setups in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "name": delegates.StringDelegate,
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda menu: menu.add(actions.CSNew),
            lambda menu: menu.add(actions.CSOpen, menu.calculation_setups,
                                  enable=bool(menu.calculation_setups)),
            lambda menu: menu.add(actions.CSDelete, menu.calculation_setups,
                                  enable=bool(menu.calculation_setups)),
            lambda menu: menu.add(actions.CSRename, menu.calculation_setups[0] if menu.single_selection else None,
                                  enable=menu.single_selection),
            lambda menu: menu.addSeparator(),
            lambda menu: menu.add(actions.CSCalculate, menu.calculation_setups[0] if menu.single_selection else None,
                                  enable=menu.single_selection),
        ]

        @property
        def calculation_setups(self):
            return [item["name"] for item in {index.internalPointer() for index in self.parent().selectedIndexes()}]

        @property
        def single_selection(self):
            return len(self.calculation_setups) == 1

    class HeaderMenu(QtWidgets.QMenu):
        """
        A header menu for the DatabasesView. Currently not used.
        """

        def __init__(self, *args, **kwargs):
            super().__init__()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        """
        Handles the mouse double click event to open the selected calculation setups.

        Args:
            event (QtGui.QMouseEvent): The mouse double click event.
        """
        if not self.selectedIndexes():
            return

        index = self.indexAt(event.pos())

        actions.CSOpen.run(index.internalPointer()["name"])


    def dragMoveEvent(self, event) -> None:
        pass

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
            for key in keys:
                act = bd.get_node(key=key)
                if act["type"] not in bd.labels.product_node_types + ["processwithreferenceproduct"]:
                    keys.remove(key)

            if not keys:
                return

            event.accept()

    def dropEvent(self, event) -> None:
        event.accept()

        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
        for key in keys:
            act = bd.get_node(key=key)
            if act["type"] not in bd.labels.product_node_types + ["processwithreferenceproduct"]:
                keys.remove(key)

        functional_units = [{key: 1.0} for key in keys]

        actions.CSNew.run(functional_units=functional_units)


class CalculationSetupsItem(widgets.ABDataItem):
    """
    An item representing a calculation setup in the tree view.
    """
    def fontData(self, col: int, key: str):
        """
        Provides font data for the item.

        Args:
            col (int): The column index.
            key (str): The key for which to provide font data.

        Returns:
            QtGui.QFont: The font data for the item.
        """
        font = super().fontData(col, key)
        if key == "name":
            font.setWeight(QtGui.QFont.Weight.DemiBold)
        return font


class CalculationSetupsModel(widgets.ABItemModel):
    """
    A model representing the data for the databases.

    Attributes:
        dataItemClass (type): The class of the data items.
    """
    dataItemClass = CalculationSetupsItem




from qtpy import QtWidgets
from qtpy.QtCore import QTimer, Slot, Signal, SignalInstance
from qtpy.QtGui import QTextFormat
from qtpy.QtWidgets import QCompleter


class ABLineEdit(QtWidgets.QLineEdit):
    textChangedDebounce: SignalInstance = Signal(str)
    _debounce_ms = 250

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._debounce_timer = QTimer(self, singleShot=True)

        self.textChanged.connect(self._set_debounce)
        self._debounce_timer.timeout.connect(self._emit_debounce)

    def _set_debounce(self):
        self._debounce_timer.setInterval(self._debounce_ms)
        self._debounce_timer.start()

    def _emit_debounce(self):
        self.textChangedDebounce.emit(self.text())

    def debounce(self):
        return self._debounce_ms

    def setDebounce(self, ms: int):
        self._debounce_ms = ms


class SignalledLineEdit(QtWidgets.QLineEdit):
    """Adapted from http://stackoverflow.com/questions/12182133/PyQt5-combine-textchanged-and-editingfinished-for-qlineedit"""

    def __init__(self, key, field, contents="", parent=None):
        super(SignalledLineEdit, self).__init__(contents, parent)
        self.editingFinished.connect(self._editing_finished)
        self.textChanged.connect(self._text_changed)
        self._before = contents
        self._key = key
        self._field = field

    @Slot(str, name="customTextChanged")
    def _text_changed(self, text: str) -> None:
        """Reset 'before' value when changing via Python"""
        if not self.hasFocus():
            self._before = text

    @Slot(name="customEditFinish")
    def _editing_finished(self) -> None:
        from activity_browser import actions

        after = self.text()
        if self._before != after:
            self._before = after
            actions.ActivityModify.run(self._key, self._field, after)


class SignalledPlainTextEdit(QtWidgets.QPlainTextEdit):
    """Adapted from https://john.nachtimwald.com/2009/08/19/better-qplaintextedit-with-line-numbers/"""

    def __init__(self, key: tuple, field: str, contents: str = "", parent=None):
        super().__init__(contents, parent)
        self.highlight()
        self.cursorPositionChanged.connect(self.highlight)
        self._before = contents
        self._key = key
        self._field = field

    @Slot(name="highlight")
    def highlight(self):
        selection = QtWidgets.QTextEdit.ExtraSelection()
        selection.format.setBackground(self.palette().alternateBase())
        selection.format.setProperty(QTextFormat.FullWidthSelection, True)
        selection.cursor = self.textCursor()
        selection.cursor.clearSelection()
        self.setExtraSelections([selection])

    def focusOutEvent(self, event):
        from activity_browser import actions

        after = self.toPlainText()
        if self._before != after:
            actions.ActivityModify.run(self._key, self._field, after)
        super().focusOutEvent(event)

    def refresh_text(self, text: str) -> None:
        self._before = text
        self.setPlainText(text)


class SignalledComboEdit(QtWidgets.QComboBox):
    """Based on SignalledPlainTextEdit.

    Could be moved to new file. Or better: this file renamed to be more inclusive
    needed to effectively implement the location dropdown list
    """

    def __init__(self, key, field, contents="", parent=None):
        super().__init__(parent)
        self._before = contents
        self._key = key
        self._field = field

    def focusOutEvent(self, event):
        from activity_browser import actions

        after = self.currentText()
        if self._before != after:
            self._before = after
            actions.ActivityModify.run(self._key, self._field, after)
        super(SignalledComboEdit, self).focusOutEvent(event)


class AutoCompleteLineEdit(QtWidgets.QLineEdit):
    """Line Edit with a completer attached"""

    def __init__(self, items: list[str], parent=None):
        super().__init__(parent=parent)
        completer = QCompleter(items, self)
        self.setCompleter(completer)



from typing import Callable, Union, List, Tuple

from qtpy import QtWidgets
from qtpy import QtCore
from qtpy.QtCore import Qt
from qtpy.QtGui import QBrush

ComboBoxItem = Union[str, Tuple[str, str], Tuple[str, str, QBrush]]


class ComboBoxDelegate(QtWidgets.QStyledItemDelegate):
    """
    A combobox delegate for use where items are scoped to a list of items
    """

    def __init__(self, item_source: Union[List[ComboBoxItem], Callable[[], list[ComboBoxItem]]], parent=None):
        """
        :param item_source: List of items to be shown in the combo box or a callable
                        returning the items, to allow the delegate to update the
                        list of items in the combobox.
                        An item can be:
                            - a string
                            - a display string - value string pair
                            - a display string, value string and text color (QBrush) triple
        """
        super(ComboBoxDelegate, self).__init__(parent)
        self.item_source = item_source  # List of items to be shown in the combo box
        self._early_commit_item_text = ""

    def set_early_commit_item(self, item_text: str):
        """Set the early commit trigger."""
        self._early_commit_item_text = item_text

    def createEditor(self, parent, option, index):
        # Get them item list
        if callable(self.item_source):
            raw_items = self.item_source()
        else:
            raw_items = self.item_source
        # Extract the colors, if provided
        item_colors = [item[2] if len(item) == 3 else None for item in raw_items]
        # Create the items in form of (str, str)
        items: list[tuple[str, str]] = []
        for item in raw_items:
            if isinstance(item, str):
                items.append((item, item))
            else:
                items.append((item[0], item[1]))

        self.item_values = [item[1] for item in items]
        editor = QtWidgets.QComboBox(parent)
        for item in items:
            userdata = None
            if not isinstance(item, str):
                item, userdata = item
            editor.addItem(item, userdata)
        editor.setCurrentIndex(self.item_values.index(index.data()))
        # Set the colors for the entries
        for i, brush in enumerate(item_colors):
            if brush is not None:
                editor.setItemData(i, brush, QtCore.Qt.ForegroundRole)
        if self._early_commit_item_text != "":
            editor.activated.connect(
                lambda index: self._handle_activated(editor, index)
            )
        return editor

    def _handle_activated(self, editor: QtWidgets.QComboBox, index: int):
        """
        In case the user selected the early commit item (either by
        clicking with the mouse, or by the arrows when the dropdown is
        closed) the selected text will be commited to the model.

        This allows the model to pop-up an editor dialog when this item
        is selected (instead of waiting for the user to commit the action
        by hitting enter or clicking away).
        """
        if editor.itemText(index) == self._early_commit_item_text:
            self.commitData.emit(editor)

    def setEditorData(self, editor, index):
        # hides the items in the background for tables views
        editor.setAutoFillBackground(True)
        current_text = index.model().data(index, Qt.EditRole)
        current_index = editor.findText(current_text)
        if current_index >= 0:
            editor.setCurrentIndex(current_index)

    def setModelData(self, editor, model, index):
        model.setData(index, self.item_values[editor.currentIndex()], Qt.EditRole)



from logging import getLogger

from qtpy import QtWidgets

import bw2data as bd

from activity_browser import application, actions
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import refresh_node
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class CSNew(ABAction):
    """
    Create a new Calculation Setup.

    This method prompts the user for a name for the new Calculation Setup (CS) if not provided.
    It validates the name to ensure it is unique within the project and creates a new CS
    with the specified functional units and impact categories.

    Args:
        name (str, optional): The name of the new Calculation Setup. If not provided, the user is prompted.
        functional_units (list[dict[tuple | int | bd.Node, float]], optional): A list of functional units to include in the CS.
        impact_categories (list[tuple], optional): A list of impact categories to include in the CS.

    Returns:
        None: Returns early if the user cancels, provides no name, or if the name already exists.

    Raises:
        None: This method does not raise exceptions but logs errors and shows warnings for invalid inputs.
    """

    icon = qicons.add
    text = "New calculation setup..."

    @staticmethod
    @exception_dialogs
    def run(name: str = None,
            functional_units: list[dict[tuple | int | bd.Node, float]] = None,
            impact_categories: list[tuple] = None
            ):

        name = name or CSNew.get_cs_name()

        # return if the user cancels or gives no name
        if not name:
            return

        # throw error if the name is already present, and return
        if name in bd.calculation_setups:
            QtWidgets.QMessageBox.warning(
                application.main_window,
                "Not possible",
                "A calculation setup with this name already exists.",
            )
            return

        inv = functional_units or []
        for i, fu in enumerate(inv):
            if not isinstance(fu, dict):
                raise TypeError("Functional units must be a list of dictionaries.")
            refreshed = {refresh_node(key).key: amount for key, amount in fu.items()}
            inv[i] = refreshed

        ia = impact_categories or []

        # instruct the CalculationSetupController to create a CS with the new name
        bd.calculation_setups[name] = {"inv": inv, "ia": ia}

        log.info(f"New calculation setup: {name}")

        actions.CSOpen.run(name)

    @staticmethod
    def get_cs_name() -> str | None:
        """
        Prompt the user for a name for the new calculation setup.
        """
        # prompt the user to give a name for the new calculation setup
        name, ok = QtWidgets.QInputDialog.getText(
            application.main_window,
            "Create new calculation setup",
            "Name of new calculation setup:" + " " * 10,
        )

        # return if the user cancels or gives no name
        if not ok or not name:
            return None
        return name



from typing import Any, List

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import uncertainty
from activity_browser.ui.icons import qicons


class ExchangeUncertaintyRemove(ABAction):
    """
    ABAction to clear the uncertainty of one or multiple exchanges.
    """

    icon = qicons.delete
    text = "Remove uncertainty/-ies"

    @staticmethod
    @exception_dialogs
    def run(exchanges: List[Any]):
        for exchange in exchanges:
            for key, value in uncertainty.EMPTY_UNCERTAINTY.items():
                exchange[key] = value

            exchange.save()



from logging import getLogger

import numpy as np
from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Signal, Slot
from stats_arrays import uncertainty_choices as uncertainty
from stats_arrays.distributions import *

from activity_browser import actions
from .. import application

from ...bwutils import PedigreeMatrix, get_uncertainty_interface
from ...bwutils.uncertainty import EMPTY_UNCERTAINTY
from ..figures import SimpleDistributionPlot

log = getLogger(__name__)


class UncertaintyWizard(QtWidgets.QWizard):
    """Using this wizard, guide the user through selecting an 'uncertainty'
    distribution (and related values) for their activity/process exchanges.

    Note that this can also be used for setting uncertainties on parameters
    """

    TYPE = 0
    PEDIGREE = 1

    complete = Signal(tuple, object)  # feed the CF uncertainty back to the origin

    def __init__(self, unc_object: object, parent=None):
        super().__init__(parent)

        self.obj = get_uncertainty_interface(unc_object)
        self.using_pedigree = False

        self.pedigree = PedigreeMatrixPage(self)
        self.type = UncertaintyTypePage(self)
        self.pages = (self.type, self.pedigree)

        for i, p in enumerate(self.pages):
            self.setPage(i, p)
        self.setStartId(self.TYPE)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.button(QtWidgets.QWizard.FinishButton).clicked.connect(
            self.update_uncertainty
        )
        self.pedigree.enable_pedigree.connect(self.used_pedigree)
        self.extract_uncertainty()

    @staticmethod
    def standard_dist_fields(dist_id: int) -> list:
        if dist_id in {2, 3}:
            return ["loc", "scale"]
        elif dist_id in {4, 7}:
            return ["minimum", "maximum"]
        elif dist_id in {5, 6}:
            return ["loc", "minimum", "maximum"]
        elif dist_id in {8, 9, 10, 11, 12}:
            return ["loc", "scale", "shape"]
        else:
            return []

    @property
    def uncertainty_info(self) -> dict:
        data = {k: v for k, v in EMPTY_UNCERTAINTY.items()}
        data["uncertainty type"] = self.field("uncertainty type")
        data["negative"] = bool(self.field("negative"))
        for field in self.standard_dist_fields(data["uncertainty type"]):
            data[field] = float(self.field(field))
        return data

    @Slot(bool, name="togglePedigree")
    def used_pedigree(self, toggle: bool) -> None:
        self.using_pedigree = toggle

    @Slot(name="modifyUncertainty")
    def update_uncertainty(self):
        """Update the uncertainty information of the relevant object, optionally
        including a pedigree update.
        """
        self.amount_mean_test()
        if self.obj.data_type == "exchange":
            actions.ExchangeModify.run(self.obj.data, self.uncertainty_info)
            if self.using_pedigree:
                actions.ExchangeModify.run(
                    self.obj.data, {"pedigree": self.pedigree.matrix.factors}
                )
        elif self.obj.data_type == "parameter":
            actions.ParameterModify.run(self.obj.data, "data", self.uncertainty_info)
            if self.using_pedigree:
                actions.ParameterModify.run(
                    self.obj.data, "data", self.pedigree.matrix.factors
                )
        elif self.obj.data_type == "cf":
            self.complete.emit(self.obj.data, self.uncertainty_info)

    def extract_uncertainty(self) -> None:
        """Used to extract possibly existing uncertainty information from the
        given exchange/parameter

        Exchange objects have uncertainty shortcuts built in, other
        objects which sometimes have uncertainty do not.
        """
        for k, v in self.obj.uncertainty.items():
            if k in EMPTY_UNCERTAINTY:
                self.setField(k, v)

        # If no loc/mean value is set yet, convert the amount.
        if not self.field("loc") or self.field("loc") == "nan":
            val = getattr(self.obj, "amount", 1.0)
            if self.field("uncertainty type") == LognormalUncertainty.id:
                val = np.log(val)
            self.setField("loc", str(val))
        # Let the other fields default to 'nan' if no values are set.
        for f in ("scale", "shape", "maximum", "minimum"):
            if not self.field(f):
                self.setField(f, "nan")

    def extract_lognormal_loc(self) -> None:
        """Special handling for looking at the uncertainty['loc'] field

        This should only be used when the 'original' set uncertainty is
        lognormal.
        """
        mean = getattr(self.obj, "amount", 1.0)
        loc = self.obj.uncertainty.get("loc", np.NaN)
        if not np.isnan(loc) and self.obj.uncertainty_type != LognormalUncertainty:
            loc = np.log(loc)
        if np.isnan(loc):
            loc = np.log(mean)
        self.setField("loc", str(loc))

    def amount_mean_test(self) -> None:
        """Asks if the 'amount' of the object should be updated to account for
        the user altering the loc/mean value.
        """
        uc_type = self.field("uncertainty type")
        no_change = {UndefinedUncertainty.id, NoUncertainty.id}
        mean = float(self.field("loc"))
        if uc_type == LognormalUncertainty.id:
            mean = np.exp(mean)
        elif uc_type in self.type.mean_is_calculated:
            mean = self.type.calculate_mean
        if not np.isclose(self.obj.amount, mean) and uc_type not in no_change:
            msg = (
                "Do you want to update the 'amount' field to match mean?"
                "\nAmount: {}\tMean: {}".format(self.obj.amount, mean)
            )
            choice = QtWidgets.QMessageBox.question(
                self,
                "Amount differs from mean",
                msg,
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.Yes,
            )
            if choice == QtWidgets.QMessageBox.Yes:
                if self.obj.data_type == "exchange":
                    actions.ExchangeModify.run(self.obj.data, {"amount": mean})

                elif self.obj.data_type == "parameter":
                    try:
                        actions.ParameterModify.run(self.obj.data, "amount", mean)
                    except Exception as e:
                        QtWidgets.QMessageBox.warning(
                            application.main_window,
                            "Could not save changes",
                            str(e),
                            QtWidgets.QMessageBox.Ok,
                            QtWidgets.QMessageBox.Ok,
                        )
                elif self.obj.data_type == "cf":
                    altered = {k: v for k, v in self.obj.uncertainty.items()}
                    altered["amount"] = mean
                    data = [*self.obj.data]
                    data[1] = altered
                    self.obj = get_uncertainty_interface(tuple(data))


class UncertaintyTypePage(QtWidgets.QWizardPage):
    """Present a list of uncertainty types directly retrieved from the `stats_arrays` package."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFinalPage(True)
        self.dist = None
        self.complete = False
        self.goto_pedigree = False
        self.previous = None
        self.mean_is_calculated = {
            TriangularUncertainty.id,
            UniformUncertainty.id,
            DiscreteUniform.id,
            BetaUncertainty.id,
        }

        # Selection of uncertainty distribution.
        box1 = QtWidgets.QGroupBox("Select the uncertainty distribution")
        self.distribution = QtWidgets.QComboBox(box1)
        self.distribution.addItems([ud.description for ud in uncertainty.choices])
        self.distribution.currentIndexChanged.connect(self.distribution_selection)
        self.registerField("uncertainty type", self.distribution, "currentIndex")
        self.pedigree = QtWidgets.QPushButton("Use pedigree")
        self.pedigree.clicked.connect(self.pedigree_page)
        box_layout = QtWidgets.QGridLayout()
        box_layout.addWidget(QtWidgets.QLabel("Distribution:"), 0, 0, 2, 1)
        box_layout.addWidget(self.distribution, 0, 1, 2, 2)
        box_layout.addWidget(self.pedigree, 0, 3, 2, 1)
        box1.setLayout(box_layout)

        # Set values for selected uncertainty distribution.
        self.field_box = QtWidgets.QGroupBox("Fill out or change required parameters")
        self.locale = QtCore.QLocale(
            QtCore.QLocale.English, QtCore.QLocale.UnitedStates
        )
        self.locale.setNumberOptions(QtCore.QLocale.RejectGroupSeparator)
        self.validator = QtGui.QDoubleValidator()
        self.validator.setLocale(self.locale)
        self.loc = QtWidgets.QLineEdit()
        self.loc.setValidator(self.validator)
        self.loc.textEdited.connect(self.balance_mean_with_loc)
        self.loc.textEdited.connect(self.check_negative)
        self.loc.textEdited.connect(self.generate_plot)
        self.loc_label = QtWidgets.QLabel("Loc:")
        self.mean = QtWidgets.QLineEdit()
        self.mean.setValidator(self.validator)
        self.mean.textEdited.connect(self.balance_loc_with_mean)
        self.mean.textEdited.connect(self.check_negative)
        self.mean.textEdited.connect(self.generate_plot)
        self.mean_label = QtWidgets.QLabel("Mean:")
        self.blocked_label = QtWidgets.QLabel("Mean:")
        self.blocked_mean = QtWidgets.QLineEdit("nan")
        self.blocked_mean.setDisabled(True)
        self.scale = QtWidgets.QLineEdit()
        self.scale.setValidator(self.validator)
        self.scale.textEdited.connect(self.generate_plot)
        self.scale_label = QtWidgets.QLabel("Sigma/scale:")
        self.shape = QtWidgets.QLineEdit()
        self.shape.setValidator(self.validator)
        self.shape.textEdited.connect(self.generate_plot)
        self.shape_label = QtWidgets.QLabel("Shape:")
        self.minimum = QtWidgets.QLineEdit()
        self.minimum.setValidator(self.validator)
        self.minimum.textEdited.connect(self.generate_plot)
        self.min_label = QtWidgets.QLabel("Minimum:")
        self.maximum = QtWidgets.QLineEdit()
        self.maximum.setValidator(self.validator)
        self.maximum.textEdited.connect(self.generate_plot)
        self.max_label = QtWidgets.QLabel("Maximum:")
        self.negative = QtWidgets.QRadioButton(self)
        self.negative.setChecked(False)
        self.negative.setHidden(True)
        box_layout = QtWidgets.QGridLayout()
        box_layout.addWidget(self.blocked_label, 0, 0)
        box_layout.addWidget(self.blocked_mean, 0, 1)
        box_layout.addWidget(self.loc_label, 2, 0)
        box_layout.addWidget(self.loc, 2, 1)
        box_layout.addWidget(self.mean_label, 2, 3)
        box_layout.addWidget(self.mean, 2, 4)
        box_layout.addWidget(self.scale_label, 4, 0)
        box_layout.addWidget(self.scale, 4, 1)
        box_layout.addWidget(self.shape_label, 6, 0)
        box_layout.addWidget(self.shape, 6, 1)
        box_layout.addWidget(self.min_label, 8, 0)
        box_layout.addWidget(self.minimum, 8, 1)
        box_layout.addWidget(self.max_label, 10, 0)
        box_layout.addWidget(self.maximum, 10, 1)
        self.field_box.setLayout(box_layout)

        self.registerField("loc", self.loc, "text")
        self.registerField("scale", self.scale, "text")
        self.registerField("shape", self.shape, "text")
        self.registerField("minimum", self.minimum, "text")
        self.registerField("maximum", self.maximum, "text")
        self.registerField("negative", self.negative, "checked")

        self.plot = SimpleDistributionPlot(self)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(box1)
        layout.addWidget(self.field_box)
        layout.addWidget(self.plot)
        self.setLayout(layout)

    def hide_param(self, *params, hide: bool = True):
        if "loc" in params:
            self.loc_label.setHidden(hide)
            self.loc.setHidden(hide)
        if "scale" in params:
            self.scale_label.setHidden(hide)
            self.scale.setHidden(hide)
        if "shape" in params:
            self.shape_label.setHidden(hide)
            self.shape.setHidden(hide)
        if "min" in params:
            self.min_label.setHidden(hide)
            self.minimum.setHidden(hide)
        if "max" in params:
            self.max_label.setHidden(hide)
            self.maximum.setHidden(hide)

    def special_distribution_handling(self):
        """Special kansas city shuffling for this distribution."""
        if self.dist.id == LognormalUncertainty.id:
            self.mean.setHidden(False)
            self.mean_label.setHidden(False)
            # Convert 'mean' to lognormal mean
            if self.previous is not None and self.previous != LognormalUncertainty.id:
                self.wizard().extract_lognormal_loc()
                self.balance_mean_with_loc()
        else:
            self.mean.setHidden(True)
            self.mean_label.setHidden(True)
            # Override the lognormal mean and copy the amount in its place
            if self.previous and self.previous == LognormalUncertainty.id:
                self.loc.setText(str(getattr(self.wizard().obj, "amount", 1)))
        # Hide or show additional untouchable 'mean' field.
        if self.dist.id in self.mean_is_calculated:
            self.blocked_label.setHidden(False)
            self.blocked_mean.setHidden(False)
        else:
            self.blocked_label.setHidden(True)
            self.blocked_mean.setHidden(True)
        self.loc_label.setText(self.distribution_loc_label)
        self.previous = self.dist.id
        self.field_box.updateGeometry()

    @property
    def distribution_loc_label(self) -> str:
        """Many distributions have a special name for the value that is entered
        into the 'loc' field.
        """
        if self.dist.id == LognormalUncertainty.id:
            return "Loc (ln(mean)):"
        elif self.dist.id == TriangularUncertainty.id:
            return "Mode:"
        elif self.dist.id == BetaUncertainty.id:
            return "Loc / alpha:"
        elif self.dist.id in {GammaUncertainty.id, WeibullUncertainty.id}:
            return "Loc / offset:"
        else:
            return "Mean:"

    @property
    def calculate_mean(self) -> float:
        """Some distributions do not specifically use a mean to generate
        their random values, in those cases present a calculated mean.

        If any of the data is missing or the calculation fails, float('nan')
        is returned.
        """
        array = self.dist.from_dicts(self.wizard().uncertainty_info)
        try:
            calc = self.dist.statistics(array).get("mean")
        # Catch exception for DiscreteUniform (https://bitbucket.org/cmutel/stats_arrays/pull-requests/5/)
        except TypeError:
            array = self.dist.fix_nan_minimum(array)
            calc = (array["maximum"] + array["minimum"]) / 2
        calc = calc.mean() if isinstance(calc, np.ndarray) else calc
        return float(calc)

    @Slot(name="changeDistribution")
    def distribution_selection(self):
        """Selected distribution and present the correct uncertainty parameters.

        See https://stats-arrays.readthedocs.io/en/latest/index.html for which
        fields to show and hide.
        """
        self.dist = uncertainty.id_dict[self.distribution.currentIndex()]

        # Huge if/elif tree to ensure the correct fields are shown.
        if self.dist.id in {0, 1}:
            self.hide_param("loc", "scale", "shape", "min", "max")
        elif self.dist.id in {2, 3}:
            self.hide_param("shape", "min", "max")
            self.hide_param("loc", "scale", hide=False)
        elif self.dist.id in {4, 7}:
            self.hide_param("loc", "scale", "shape")
            self.hide_param("min", "max", hide=False)
        elif self.dist.id in {5, 6}:
            self.hide_param("scale", "shape")
            self.hide_param("loc", "min", "max", hide=False)
        elif self.dist.id in {8, 9, 10, 11, 12}:
            self.hide_param("min", "max")
            self.hide_param("loc", "scale", "shape", hide=False)
        self.special_distribution_handling()
        self.generate_plot()

    def completed_active_fields(self) -> bool:
        """Returns a boolean value based on the distribution id.
        If the distribution contains an average, minimum and maximum this forces the
        average to exist exclusively within these bounds"""
        completed = False
        if self.dist.id in {0, 1}:
            completed = True
        elif self.dist.id in {2, 3}:
            completed = all(
                [
                    field.hasAcceptableInput() and field.text()
                    for field in (self.loc, self.scale)
                ]
            )
        elif self.dist.id in {4, 7}:
            completed = all(
                [
                    field.hasAcceptableInput() and field.text()
                    for field in (self.minimum, self.maximum)
                ]
            )
        elif self.dist.id in {5, 6}:
            completed = all(
                [
                    field.hasAcceptableInput() and field.text()
                    for field in (self.minimum, self.maximum, self.loc)
                ]
            ) and (
                float(self.minimum.text())
                < float(self.loc.text())
                < float(self.maximum.text())
            )
        elif self.dist.id in {8, 9, 10, 11, 12}:
            completed = all(
                [
                    field.hasAcceptableInput() and field.text()
                    for field in (self.scale, self.shape, self.loc)
                ]
            )
        return completed

    @Slot(name="locToMean")
    def balance_mean_with_loc(self):
        if self.loc.text():
            self.mean.setText(str(np.exp(float(self.loc.text()))))

    @Slot(name="meanToLoc")
    def balance_loc_with_mean(self):
        if not self.mean.hasAcceptableInput():
            self.loc.setText("nan")
            return
        val = float(self.mean.text() if self.mean.text() else "nan")
        val = -1 * val if val < 0 else val
        self.loc.setText(str(np.log(val) if val != 0 else float("nan")))

    @Slot(name="testValueNegative")
    def check_negative(self) -> None:
        """Determine which QLineEdit to use to set the negative value.

        Another special edge-case for the lognormal distribution.
        """
        if not self.mean.hasAcceptableInput():
            return
        val = float(self.mean.text() if self.mean.text() else "nan")
        if self.dist.id == LognormalUncertainty.id and val < 0:
            self.setField("negative", True)
        else:
            self.setField("negative", False)

    def initializePage(self) -> None:
        self.distribution_selection()
        self.balance_mean_with_loc()

    def nextId(self) -> int:
        if self.goto_pedigree:
            return UncertaintyWizard.PEDIGREE
        return -1

    def isComplete(self) -> bool:
        return self.complete

    @Slot(name="gotoPedigreePage")
    def pedigree_page(self) -> None:
        self.goto_pedigree = True
        self.wizard().next()

    @Slot(name="regenPlot")
    def generate_plot(self) -> None:
        """Called whenever a value changes, (re)generate the plot.

        Also tests if all of the visible QLineEdit fields have valid values.
        """
        self.complete = self.completed_active_fields()
        no_dist = self.dist.id in {UndefinedUncertainty.id, NoUncertainty.id}
        if self.complete or no_dist:
            array = self.dist.from_dicts(self.wizard().uncertainty_info)
            if self.dist.id in self.mean_is_calculated:
                mean = self.calculate_mean
                self.blocked_mean.setText(str(mean))
            if self.dist.id == LognormalUncertainty.id:
                mean = self.dist.statistics(array).get("median")
            elif no_dist:
                mean = self.wizard().obj.amount
            else:
                mean = self.dist.statistics(array).get("mean")
            data = self.dist.random_variables(array, 1000)
            if not np.any(np.isnan(data)):
                self.plot.plot(data, mean)
        self.completeChanged.emit()


class PedigreeMatrixPage(QtWidgets.QWizardPage):
    """Guide the user through filling out a pedigree matrix.

    There are 5 indicators used, each carrying a score from 1 to 5
    with 1 indicating 'less uncertain' and 5 'more uncertain'.

    NOTE: Currently, the pedigree matrix will always default to a lognormal distribution.

    NOTE: using terms and quoting from the paper:
    'Empirically based uncertainty factors for the pedigree matrix in ecoinvent' (2016)
    doi: 10.1007/s11367-013-0670-5
    """

    enable_pedigree = Signal(bool)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFinalPage(True)
        self.matrix = None

        self.field_box = QtWidgets.QGroupBox("Fill out or change required parameters")
        self.locale = QtCore.QLocale(
            QtCore.QLocale.English, QtCore.QLocale.UnitedStates
        )
        self.locale.setNumberOptions(QtCore.QLocale.RejectGroupSeparator)
        self.validator = QtGui.QDoubleValidator()
        self.validator.setLocale(self.locale)
        self.loc = QtWidgets.QLineEdit()
        self.loc.setValidator(self.validator)
        self.loc.textEdited.connect(self.balance_mean_with_loc)
        self.loc.textEdited.connect(self.check_negative)
        self.loc.textEdited.connect(self.check_complete)
        self.mean = QtWidgets.QLineEdit()
        self.mean.setValidator(self.validator)
        self.mean.textEdited.connect(self.balance_loc_with_mean)
        self.mean.textEdited.connect(self.check_negative)
        self.mean.textEdited.connect(self.check_complete)
        box_layout = QtWidgets.QGridLayout()
        box_layout.addWidget(QtWidgets.QLabel("Loc (ln(mean)):"), 0, 0)
        box_layout.addWidget(self.loc, 0, 1)
        box_layout.addWidget(QtWidgets.QLabel("Mean:"), 0, 3)
        box_layout.addWidget(self.mean, 0, 4)
        self.field_box.setLayout(box_layout)

        box = QtWidgets.QGroupBox("Select pedigree values")

        self.reliable = QtWidgets.QComboBox(box)
        self.reliable.addItems(
            [
                "1) Verified data based on measurements",
                "2) Verified data partly based on assumptions",
                "3) Non-verified data partly based on qualified measurements",
                "4) Qualified estimate",
                "5) Non-qualified estimate",
            ]
        )
        self.complete = QtWidgets.QComboBox(box)
        self.complete.addItems(
            [
                "1) Representative relevant data from all sites, over an adequate period",
                "2) Representative relevant data from >50% sites, over an adequate period",
                "3) Representative relevant data from <50% sites OR >50%, but over shorter period",
                "4) Representative relevant data from one site OR some sites but over shorter period",
                "5) Representativeness unknown",
            ]
        )
        self.temporal = QtWidgets.QComboBox(box)
        self.temporal.addItems(
            [
                "1) Data less than 3 years old",
                "2) Data less than 6 years old",
                "3) Data less than 10 years old",
                "4) Data less than 15 years old",
                "5) Data age unknown or more than 15 years old",
            ]
        )
        self.geographical = QtWidgets.QComboBox(box)
        self.geographical.addItems(
            [
                "1) Data from area under study",
                "2) Average data from larger area in which area under study is included",
                "3) Data from area with similar production conditions",
                "4) Data from area with slightly similar production conditions",
                "5) Data from unknown OR distinctly different area",
            ]
        )
        self.technological = QtWidgets.QComboBox(box)
        self.technological.addItems(
            [
                "1) Data from enterprises, processes and materials under study",
                "2) Data from processes and materials under study, different enterprise",
                "3) Data from processes and materials under study from different technology",
                "4) Data on related processes and materials",
                "5) Data on related processes on lab scale OR from different technology",
            ]
        )
        self.reliable.currentIndexChanged.connect(self.check_complete)
        self.complete.currentIndexChanged.connect(self.check_complete)
        self.temporal.currentIndexChanged.connect(self.check_complete)
        self.geographical.currentIndexChanged.connect(self.check_complete)
        self.technological.currentIndexChanged.connect(self.check_complete)

        box_layout = QtWidgets.QGridLayout()
        box_layout.addWidget(QtWidgets.QLabel("Reliability"), 0, 0, 2, 2)
        box_layout.addWidget(self.reliable, 0, 2, 2, 3)
        box_layout.addWidget(QtWidgets.QLabel("Completeness"), 2, 0, 2, 2)
        box_layout.addWidget(self.complete, 2, 2, 2, 3)
        box_layout.addWidget(QtWidgets.QLabel("Temporal correlation"), 4, 0, 2, 2)
        box_layout.addWidget(self.temporal, 4, 2, 2, 3)
        box_layout.addWidget(QtWidgets.QLabel("Geographical correlation"), 6, 0, 2, 2)
        box_layout.addWidget(self.geographical, 6, 2, 2, 3)
        box_layout.addWidget(
            QtWidgets.QLabel("Further technological correlation"), 8, 0, 2, 2
        )
        box_layout.addWidget(self.technological, 8, 2, 2, 3)
        box.setLayout(box_layout)

        self.plot = SimpleDistributionPlot(self)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.field_box)
        layout.addWidget(box)
        layout.addWidget(self.plot)
        self.setLayout(layout)

    def cleanupPage(self):
        self.enable_pedigree.emit(False)

    def initializePage(self):
        # if the parent contains an 'obj' with uncertainty, extract data
        self.setField("uncertainty type", 2)
        self.loc.setText(self.field("loc"))
        self.balance_mean_with_loc()
        obj = getattr(self.wizard(), "obj")
        try:
            matrix = PedigreeMatrix.from_dict(obj.uncertainty.get("pedigree", {}))
            self.pedigree = matrix.factors
        except AssertionError as e:
            log.info("Could not extract pedigree data: {}".format(str(e)))
            self.pedigree = {}
        self.check_complete()

    def nextId(self):
        """Ensures that 'Next' button does not show."""
        return -1

    @property
    def pedigree(self) -> tuple:
        return (
            self.reliable.currentIndex() + 1,
            self.complete.currentIndex() + 1,
            self.temporal.currentIndex() + 1,
            self.geographical.currentIndex() + 1,
            self.technological.currentIndex() + 1,
        )

    @pedigree.setter
    def pedigree(self, data: dict) -> None:
        self.reliable.setCurrentIndex(data.get("reliability", 1) - 1)
        self.complete.setCurrentIndex(data.get("completeness", 1) - 1)
        self.temporal.setCurrentIndex(data.get("temporal correlation", 1) - 1)
        self.geographical.setCurrentIndex(data.get("geographical correlation", 1) - 1)
        self.technological.setCurrentIndex(
            data.get("further technological correlation", 1) - 1
        )

    @Slot(name="locToMean")
    def balance_mean_with_loc(self):
        self.setField("loc", self.loc.text())
        if self.loc.text():
            self.mean.setText(str(np.exp(float(self.loc.text()))))

    @Slot(name="meanToLoc")
    def balance_loc_with_mean(self):
        if not self.mean.hasAcceptableInput():
            self.loc.setText("nan")
            return
        val = float(self.mean.text() if self.mean.text() else "nan")
        val = -1 * val if val < 0 else val
        loc_val = str(np.log(val)) if val != 0 else "nan"
        self.loc.setText(loc_val)
        self.setField("loc", loc_val)

    @Slot(name="testValueNegative")
    def check_negative(self) -> None:
        """Determine which QLineEdit to use to set the negative value.

        Another special edge-case for the lognormal distribution.
        """
        if not self.mean.hasAcceptableInput():
            return
        val = float(self.mean.text() if self.mean.text() else "nan")
        if val < 0:
            self.setField("negative", True)
        else:
            self.setField("negative", False)

    @Slot(name="constructPedigreeMatrix")
    def check_complete(self) -> None:
        self.matrix = PedigreeMatrix.from_numbers(self.pedigree)
        self.setField("scale", self.matrix.calculate())
        self.generate_plot()

    @Slot(name="regenPlot")
    def generate_plot(self) -> None:
        """Called whenever a value changes, (re)generate the plot.

        Also tests if all of the visible QLineEdit fields have valid values.
        """
        array = LognormalUncertainty.from_dicts(self.wizard().uncertainty_info)
        median = LognormalUncertainty.statistics(array).get("median")
        data = LognormalUncertainty.random_variables(array, 1000)
        if not np.any(np.isnan(data)):
            self.plot.plot(data, median)
        self.enable_pedigree.emit(True)



from typing import List
from logging import getLogger

from qtpy import QtWidgets

import bw2data as bd

from activity_browser import signals
from activity_browser.ui import widgets
from activity_browser.actions.base import ABAction, exception_dialogs

log = getLogger(__name__)


class MethodRename(ABAction):
    """
    Renames an existing method in the Brightway2 data structure.

    This method allows renaming a single method by prompting the user for a new name.
    It ensures that only one method is renamed at a time, validates the new name, and
    updates the method in the Brightway2 database.

    Args:
        method_name (tuple[str] | list[tuple[str]]): The name of the method to rename.
            If a list is provided, it must contain exactly one tuple.

    Steps:
    - Ensure only one method is being renamed at a time.
    - Check if the method exists in the Brightway2 database.
    - Open a dialog to prompt the user for the new method name.
    - Validate the new name to ensure it is not empty and does not already exist.
    - Copy the method to the new name and process it.
    - Emit a signal to notify that the method has been renamed.
    - Deregister the old method.

    Raises:
        ValueError: If more than one method is provided for renaming.
        RuntimeError: If the method does not exist, the new name is empty, or the new name already exists.
    """

    text = "Rename Impact Category"

    @staticmethod
    @exception_dialogs
    def run(method_name: tuple[str] | list[tuple[str]]):
        # safeguard: only allow renaming one method at a time
        if isinstance(method_name, list):
            if len(method_name) != 1 or not isinstance(method_name[0], tuple):
                raise ValueError("Can only rename one method at a time.")
            method_name = method_name[0]

        # check if method exists
        if method_name not in bd.methods:
            raise RuntimeError(f"Method {method_name} does not exist.")

        method = bd.Method(method_name)

        # open dialog to get new name
        dialog = widgets.ABListEditDialog(method_name)
        dialog.exec_()

        # if dialog was cancelled, do nothing
        if not dialog.result() == QtWidgets.QDialog.Accepted:
            return

        new_name = dialog.get_data(as_tuple=True)

        # check new name validity
        if len(new_name) == 0:
            raise RuntimeError("Method name cannot be empty.")

        if new_name in bd.methods:
            raise RuntimeError(f"Method {new_name} already exists.")

        # copy method to new name and process
        method.copy(new_name).process()

        # this should not happen like this, as the model and therefore signals should be handled declaritavely,
        # but since method renaming is not native to bw2data we have to do it manually here
        signals.method.renamed.emit(method_name, new_name)

        # deregister old method
        method.deregister()



from qtpy import QtWidgets

import pandas as pd
import bw2data as bd
import bw_functional as bf

from activity_browser import actions, bwutils
from activity_browser.ui import widgets, icons


class ConsumersTab(QtWidgets.QWidget):
    """
    A widget that displays consumers related to a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display consumers for.
        view (ConsumersView): The view displaying the consumers.
        model (ConsumersModel): The model containing the data for the consumers.
    """
    def __init__(self, activity: tuple | int | bd.Node, parent=None):
        """
        Initializes the ConsumersTab widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display consumers for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)

        self.activity = bwutils.refresh_node(activity)

        self.view = ConsumersView(self)
        self.model = ConsumersModel(self)
        self.view.setModel(self.model)
        self.view.setSortingEnabled(True)

        self.build_layout()
        self.sync()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 10, 0, 1)
        layout.addWidget(self.view)
        self.setLayout(layout)

    def sync(self):
        """
        Synchronizes the widget with the current state of the activity.
        """
        self.activity = bwutils.refresh_node(self.activity)
        exchanges = []
        if isinstance(self.activity, bf.Process):
            for product in self.activity.products():
                exchanges += list(product.upstream())
        else:
            exchanges = list(self.activity.upstream())

        self.model.setDataFrame(self.build_df(exchanges))

    def build_df(self, exchanges: list[bd.Edge]) -> pd.DataFrame:
        """
        Builds a DataFrame from the given exchanges.

        Args:
            exchanges (list): The list of exchanges to build the DataFrame from.

        Returns:
            pd.DataFrame: The DataFrame containing the exchanges data.
        """
        exc_df = pd.DataFrame(exchanges, columns=["amount", "input", "output"])
        input_df = bwutils.AB_metadata.get_metadata(exc_df["input"].unique(), ["name", "type", "unit", "key"])
        output_df = bwutils.AB_metadata.get_metadata(exc_df["output"].unique(), ["name", "type", "key"])

        df = exc_df.merge(
            input_df.rename({"name": "product", "type": "_product_type"}, axis="columns"),
            left_on="input",
            right_on="key",
        ).drop(columns=["key"])

        df = df.merge(
            output_df.rename({"name": "consumer", "type": "_consumer_type"}, axis="columns"),
            left_on="output",
            right_on="key",
        ).drop(columns=["key"])

        df = df.rename({"input": "_product_key", "output": "_consumer_key"}, axis="columns")

        cols = ["amount", "unit", "product", "consumer"]
        cols += [col for col in df.columns if col.startswith("_")]

        return df[cols]


class ConsumersView(widgets.ABTreeView):
    """
    A view that displays the consumers in a tree structure.
    """
    def mouseDoubleClickEvent(self, event) -> None:
        """
        Handles the mouse double-click event.

        Args:
            event: The mouse event.
        """
        items = [i.internalPointer() for i in self.selectedIndexes() if isinstance(i.internalPointer(), ConsumersItem)]
        keys = list({i["_consumer_key"] for i in items})
        if keys:
            actions.ActivityOpen.run(keys)


class ConsumersItem(widgets.ABDataItem):
    """
    An item representing a consumer in the tree view.
    """
    def decorationData(self, col, key):
        """
        Provides decoration data for the item.

        Args:
            col: The column index.
            key: The key for which to provide decoration data.

        Returns:
            The decoration data for the item.
        """
        if key not in ["product", "consumer"]:
            return

        if key == "product":
            activity_type = self["_product_type"]
        else:  # key is "consumer"
            activity_type = self["_consumer_type"]

        if activity_type in ["natural resource", "emission", "inventory indicator", "economic", "social"]:
            return icons.qicons.biosphere
        if activity_type == "processwithreferenceproduct":
            return icons.qicons.processproduct
        if activity_type == "product":
            return icons.qicons.product
        if activity_type in ["process", "multifunctional", "nonfunctional"]:
            return icons.qicons.process
        if activity_type == "waste":
            return icons.qicons.waste


class ConsumersModel(widgets.ABItemModel):
    """
    A model representing the data for the consumers.

    Attributes:
        dataItemClass (type): The class of the data items.
    """
    dataItemClass = ConsumersItem



from typing import Any, List

from bw2data.parameters import ParameterizedExchange

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class ExchangeFormulaRemove(ABAction):
    """
    ABAction to clear the formula's of one or more exchanges.
    """

    icon = qicons.delete
    text = "Clear formula(s)"

    @staticmethod
    @exception_dialogs
    def run(exchanges: List[Any]):
        for exchange in exchanges:
            if "formula" not in exchange:
                return

            del exchange["formula"]
            exchange.save()

            ParameterizedExchange.delete().where(ParameterizedExchange.exchange == exchange._document.id).execute()



import pandas as pd
from qtpy import QtWidgets

from activity_browser import application, signals
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import AB_metadata, commontasks
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class ActivityDuplicateToLoc(ABAction):
    """
    ABAction to duplicate an activity and possibly their exchanges to a new location.
    """

    icon = qicons.copy
    text = "Duplicate node to new location"

    @classmethod
    @exception_dialogs
    def run(cls, activity_key: tuple):
        activity = bd.get_activity(activity_key)
        db_name = activity["database"]

        # get list of dependent databases for activity and load to MetaDataStore
        databases = []
        for exchange in activity.technosphere():
            databases.append(exchange.input[0])
        if db_name not in databases:  # add own database if it wasn't added already
            databases.append(db_name)

        # load all dependent databases to MetaDataStore
        dbs = {db: AB_metadata.get_database_metadata(db) for db in databases}
        # get list of all unique locations in the dependent databases (sorted alphabetically)
        locations = []
        for db in dbs.values():
            locations += db["location"].to_list()  # add all locations to one list
        locations = list(set(locations))  # reduce the list to only unique items
        locations.sort()

        # get the location to relink
        db = dbs[db_name]
        old_location = db.loc[db["key"] == activity.key]["location"].iloc[0]

        # trigger dialog with autocomplete-writeable-dropdown-list
        options = (old_location, locations)
        dialog = LocationLinkingDialog.relink_location(
            activity["name"], options, application.main_window
        )

        if dialog.exec_() != LocationLinkingDialog.Accepted:
            return

        # read the data from the dialog
        for old, new in dialog.relink.items():
            alternatives = []
            new_location = new
            if dialog.use_rer.isChecked():  # RER
                alternatives.append(dialog.use_rer.text())
            if dialog.use_ews.isChecked():  # Europe without Switzerland
                alternatives.append(dialog.use_ews.text())
            if dialog.use_row.isChecked():  # RoW
                alternatives.append(dialog.use_row.text())
            # the order we add alternatives is important, they are checked in this order!
            if len(alternatives) > 0:
                use_alternatives = True
            else:
                use_alternatives = False

        # successful_links = {}  # dict of dicts, key of new exch : {new values} <-- see 'values' below
        # in the future, 'alternatives' could be improved by making use of some location hierarchy. From that we could
        # get things like if the new location is NL but there is no NL, but RER exists, we use that. However, for that
        # we need some hierarchical structure to the location data, which may be available from ecoinvent, but we need
        # to look for that.

        # get exchanges that we want to relink
        # for exch in activity.technosphere():
        #     candidate = self.find_candidate(dbs, exch, old_location, new_location, use_alternatives, alternatives)
        #     if candidate is None:
        #         continue  # no suitable candidate was found, try the next exchange
        #
        #     # at this point, we have found 1 suitable candidate, whether that is new_location or alternative location
        #     values = {
        #         'amount': exch.get('amount', False),
        #         'comment': exch.get('comment', False),
        #         'formula': exch.get('formula', False),
        #         'uncertainty': exch.get('uncertainty', False)
        #     }
        #     successful_links[candidate['key'].iloc[0]] = values

        # now, create a new activity by copying the old one
        new_code = commontasks.generate_copy_code(activity.key)
        new_act = activity.copy(new_code)

        # update production exchanges
        # TODO: check if this is even necessary (I think BW takes care of this)
        for exc in new_act.production():
            if exc.input.key == activity.key:
                exc.input = new_act
                exc.save()

        # update 'products'
        for product in new_act.get("products", []):
            if product.get("input") == activity.key:
                product.input = new_act.key

        # save the new location to the activity
        new_act["location"] = new_location

        new_act.save()

        # get exchanges that we want to delete
        # del_exch = []  # delete these exchanges
        for exch in new_act.technosphere():
            candidate = cls.find_candidate(
                db_name,
                dbs,
                exch,
                old_location,
                new_location,
                use_alternatives,
                alternatives,
            )
            if candidate is None:
                continue  # no suitable candidate was found, try the next exchange
            exch.input = candidate["key"][0]
            exch.save()
            # del_exch.append(exch)
        # delete exchanges with old locations
        # exchange_controller.delete_exchanges(del_exch)

        # add the new exchanges with all values carried over from last exchange
        # exchange_controller.add_exchanges(list(successful_links.keys()), new_act.key, successful_links)

        # update the MetaDataStore and open new activity
        AB_metadata.update_metadata(new_act.key)
        signals.safe_open_activity_tab.emit(new_act.key)

    @staticmethod
    def find_candidate(
        db_name, dbs, exch, old_location, new_location, use_alternatives, alternatives
    ):
        """Find a candidate to replace the exchange with."""
        current_db = exch.input[0]
        if current_db == db_name:
            db = dbs[current_db]
        else:  # if the exchange is not from the current database, also check the current
            # (user may have added their own alternative dependents already)
            db = pd.concat([dbs[current_db], dbs[db_name]])

        if db.loc[db["key"] == exch.input]["location"].iloc[0] != old_location:
            return  # this exchange has a location we're not trying to re-link

        # get relevant data to match on
        row = db.loc[db["key"] == exch.input]
        name = row["name"].iloc[0]
        prod = row["reference product"].iloc[0]
        unit = row["unit"].iloc[0]

        # get candidates to match (must have same name, product and unit)
        candidates = db.loc[
            (db["name"] == name)
            & (db["reference product"] == prod)
            & (db["unit"] == unit)
        ]
        if len(candidates) <= 1:
            return  # this activity does not exist in this database with another location (1 is self)

        # check candidates for new_location
        candidate = candidates.loc[candidates["location"] == new_location]
        if len(candidate) == 0 and not use_alternatives:
            return  # there is no candidate
        elif len(candidate) > 1:
            return  # there is more than one candidate, we can't know what to use
        elif len(candidate) == 0:
            # there are no candidates, but we can try alternatives
            for alt in alternatives:
                candidate = candidates.loc[candidates["location"] == alt]
                if len(candidate) == 1:
                    break  # found an alternative in with this alternative location, stop looking
            if len(candidate) != 1:
                return  # there are either no or multiple matches with alternative locations
        return candidate


class LocationLinkingDialog(QtWidgets.QDialog):
    """Display all of the possible location links in a single dialog for the user.

    Allow users to select alternate location links and an option to link to generic alternatives (GLO, RoW).
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Activity Location linking")

        self.loc_label = QtWidgets.QLabel()
        self.label_choices = []
        self.grid_box = QtWidgets.QGroupBox("Location link:")
        self.grid = QtWidgets.QGridLayout()
        self.grid_box.setLayout(self.grid)

        self.use_alternatives_label = QtWidgets.QLabel(
            "Use generic alternatives as fallback:"
        )
        self.use_alternatives_label.setToolTip(
            "If the chosen location is not found, try matching the selected "
            "locations below too"
        )
        self.use_row = QtWidgets.QCheckBox("RoW")
        self.use_row.setChecked(True)
        self.use_rer = QtWidgets.QCheckBox("RER")
        self.use_ews = QtWidgets.QCheckBox("Europe without Switzerland")

        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
        )
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.loc_label)
        layout.addWidget(self.grid_box)
        layout.addWidget(self.use_alternatives_label)
        layout.addWidget(self.use_row)
        layout.addWidget(self.use_rer)
        layout.addWidget(self.use_ews)
        layout.addWidget(self.buttons)
        self.setLayout(layout)

    @property
    def relink(self) -> dict:
        """Returns a dictionary of str -> str key/values, showing which keys
        should be linked to which values.

        Only returns key/value pairs if they differ.
        """
        return {
            label.text(): combo.currentText()
            for label, combo in self.label_choices
            if label.text() != combo.currentText()
        }

    @classmethod
    def construct_dialog(
        cls,
        label: str,
        options: list,
        parent: QtWidgets.QWidget = None,
    ) -> "LocationLinkingDialog":
        loc, locs = options

        obj = cls(parent)
        obj.loc_label.setText(label)

        label = QtWidgets.QLabel(loc)
        combo = QtWidgets.QComboBox()
        combo.addItems(locs)
        combo.setCurrentText(loc)
        obj.label_choices.append((label, combo))
        # Start at 1 because row 0 is taken up by the loc_label
        obj.grid.addWidget(label, 0, 0, 1, 2)
        obj.grid.addWidget(combo, 0, 2, 1, 2)

        obj.updateGeometry()
        return obj

    @classmethod
    def relink_location(
        cls, act_name: str, options: list, parent=None
    ) -> "LocationLinkingDialog":
        label = "Relinking exchanges from activity '{}' to a new location.".format(
            act_name
        )
        return cls.construct_dialog(label, options, parent)





# -*- coding: utf-8 -*-
import math

from qtpy.QtWidgets import QStyledItemDelegate

from .float import FloatDelegate
from .uncertainty import UncertaintyDelegate


class ViewOnlyDelegate(QStyledItemDelegate):
    """Disable the editor functionality to allow specific columns of an
    editable table to be view-only.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

    def displayText(self, value, locale):
        try:
            value = float(value)
            if math.isnan(value):
                return ""
            return "{:.5g}".format(value)
        except ValueError:
            return str(value)
        except TypeError:
            # Unexpected object types (eg. parameters)
            return repr(value)

    def createEditor(self, parent, option, index):
        return None


class ViewOnlyFloatDelegate(FloatDelegate):
    """Correctly display float values without allowing modification."""

    def createEditor(self, parent, option, index):
        return None


class ViewOnlyUncertaintyDelegate(UncertaintyDelegate):
    """Correctly display uncertainty type without allowing modification."""

    def createEditor(self, parent, option, index):
        return None



import os.path
from logging import getLogger

from qtpy.QtCore import Signal, SignalInstance

from activity_browser import application
from activity_browser.actions.base import exception_dialogs
from activity_browser.ui import icons, widgets, threading
from activity_browser.bwutils.io.ecoinvent_lcia_importer import EcoinventLCIAImporter

from .method_importer_ecoinvent import ExtractExcelThread, MethodImporterEcoinvent

log = getLogger(__name__)


class MethodImporterBW2IO(MethodImporterEcoinvent):
    """ABAction to import ecoinvent methods shipped with BW2IO"""

    icon = icons.qicons.import_db
    text = "Import methods from BW2IO"
    tool_tip = "Import methods that come shipped with BW2IO"

    @classmethod
    @exception_dialogs
    def run(cls):
        # initialize the import thread, setting needed attributes
        extract_thread = ExtractMethodsThread(application)
        extract_thread.loaded.connect(cls.write_database)

        # show progress dialog for importing the excel
        progress_dialog = widgets.ABProgressDialog.get_connected_dialog("Importing Database")
        extract_thread.finished.connect(progress_dialog.deleteLater)

        extract_thread.start()


class ExtractMethodsThread(threading.ABThread):
    loaded: SignalInstance = Signal(EcoinventLCIAImporter)

    def run_safely(self):
        import zipfile
        import json
        from bw2io.data import dirpath

        fp = os.path.join(dirpath, "lcia", "lcia_39_ecoinvent.zip")

        with zipfile.ZipFile(fp, mode="r") as archive:
            data = json.load(archive.open("data.json"))

        for method in data:
            method['name'] = tuple(method['name'])
            for obj in method['exchanges']:
                del obj['input']

        ei = EcoinventLCIAImporter("lcia_39_ecoinvent.zip")
        ei.data = data
        self.loaded.emit(ei)




from asteval import Interpreter
from qtpy.QtCore import Slot
from qtpy.QtGui import QContextMenuEvent, QDragMoveEvent, QDropEvent
from qtpy.QtWidgets import QAction, QMenu

import bw2data as bd
import bw_functional as bf

from activity_browser import actions, signals
from activity_browser.ui import icons, delegates

from .parameter_models import (
    BaseParameterModel,
    ProjectParameterModel,
    DatabaseParameterModel,
    ActivityParameterModel,
    ParameterTreeModel,
    ScenarioModel
)
from .base import ABDataFrameView, ABDictTreeView


class ScenarioTable(ABDataFrameView):
    """Constructs an infinitely (horizontally) expandable table that is
    used to set specific amount for user-defined parameters.

    The two required columns in the dataframe for the table are 'Name',
    and 'Type'. all other columns are seen as scenarios containing N floats,
    where N is the number of rows found in the Name column.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.table_name = "scenario_table"

        self.horizontalHeader().setStretchLastSection(False)
        self.verticalHeader().setVisible(True)

        self.model = ScenarioModel(self)
        self.model.updated.connect(self.update_proxy_model)
        signals.project.changed.connect(self.group_column)

    @Slot(bool, name="showGroupColumn")
    def group_column(self, shown: bool = False) -> None:
        self.setColumnHidden(0, not shown)

    def iterate_scenarios(self) -> list[tuple[str, list]]:
        return self.model.iterate_scenarios()


class BaseParameterTable(ABDataFrameView):
    MODEL = BaseParameterModel

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSelectionMode(ABDataFrameView.SingleSelection)

        self.model = self.MODEL(self)
        self.doubleClicked.connect(
            lambda: self.model.handle_double_click(self.currentIndex())
        )
        self.delete_action = QAction(icons.qicons.delete, "Delete parameter", None)
        self.delete_action.triggered.connect(
            lambda: self.model.delete_parameter(self.currentIndex())
        )
        self.rename_action = QAction(icons.qicons.edit, "Rename parameter", None)
        self.rename_action.triggered.connect(
            lambda: self.model.handle_parameter_rename(self.currentIndex())
        )
        self.modify_uncertainty_action = QAction(
            icons.qicons.edit, "Modify uncertainty", None
        )
        self.modify_uncertainty_action.triggered.connect(self.modify_uncertainty)
        self.remove_uncertainty_action = QAction(
            icons.qicons.delete, "Remove uncertainty", None
        )
        self.remove_uncertainty_action.triggered.connect(self.remove_uncertainty)
        self.model.updated.connect(self.update_proxy_model)

        # hide raw parameter column
        self.model.updated.connect(
            lambda: self.setColumnHidden(self.model.param_col, True)
        )
        self.model.updated.connect(lambda: self.resizeColumnToContents(0))

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        """Have the parameter test to see if it can be deleted safely."""
        if self.indexAt(event.pos()).row() == -1:
            return
        menu = QMenu(self)
        menu.addAction(self.rename_action)
        menu.addAction(self.modify_uncertainty_action)
        menu.addSeparator()
        menu.addAction(self.delete_action)
        menu.addAction(self.remove_uncertainty_action)
        proxy = self.indexAt(event.pos())
        if proxy.isValid():
            param = self.get_parameter(proxy)
            if param.is_deletable():
                self.delete_action.setEnabled(True)
            else:
                self.delete_action.setEnabled(False)
            menu.exec_(event.globalPos())

    def get_parameter(self, proxy):
        return self.model.get_parameter(proxy)

    def get_key(self, *args) -> tuple:
        return self.model.get_key()

    def delete_parameter(self, proxy) -> None:
        self.model.delete_parameter(proxy)

    @Slot(name="modifyParameterUncertainty")
    def modify_uncertainty(self) -> None:
        proxy = next(p for p in self.selectedIndexes())
        self.model.modify_uncertainty(proxy)

    @Slot(name="unsetParameterUncertainty")
    def remove_uncertainty(self) -> None:
        proxy = next(p for p in self.selectedIndexes())
        self.model.remove_uncertainty(proxy)

    def comment_column(self, show: bool):
        self.setColumnHidden(self.model.comment_col, not show)
        self.resizeColumnsToContents()
        self.resizeRowsToContents()


class ProjectParameterTable(BaseParameterTable):
    MODEL = ProjectParameterModel

    def __init__(self, parent=None):
        super().__init__(parent)
        self.table_name = "project_parameter"

        # Set delegates for specific columns
        self.setItemDelegateForColumn(1, delegates.FloatDelegate(self))
        self.setItemDelegateForColumn(2, delegates.FormulaDelegate(self))
        self.setItemDelegateForColumn(3, delegates.StringDelegate(self))
        self.setItemDelegateForColumn(4, delegates.ViewOnlyUncertaintyDelegate(self))

    def uncertainty_columns(self, show: bool):
        for i in range(4, 10):
            self.setColumnHidden(i, not show)

    @staticmethod
    def get_usable_parameters():
        return ProjectParameterModel.get_usable_parameters()

    @staticmethod
    def get_interpreter() -> Interpreter:
        return ProjectParameterModel.get_interpreter()


class DataBaseParameterTable(BaseParameterTable):
    MODEL = DatabaseParameterModel

    def __init__(self, parent=None):
        super().__init__(parent)
        self.table_name = "database_parameter"

        # Set delegates for specific columns
        self.setItemDelegateForColumn(1, delegates.FloatDelegate(self))
        self.setItemDelegateForColumn(2, delegates.FormulaDelegate(self))
        self.setItemDelegateForColumn(3, delegates.DatabaseDelegate(self))
        self.setItemDelegateForColumn(4, delegates.StringDelegate(self))
        self.setItemDelegateForColumn(5, delegates.ViewOnlyUncertaintyDelegate(self))

    def uncertainty_columns(self, show: bool):
        for i in range(5, 11):
            self.setColumnHidden(i, not show)

    def get_key(self) -> tuple:
        return self.model.get_key(self.currentIndex())

    @staticmethod
    def get_usable_parameters():
        return DatabaseParameterModel.get_usable_parameters()

    def get_interpreter(self) -> Interpreter:
        """Take the interpreter from the ProjectParameterTable and add
        (potentially overwriting) all database symbols for the selected index.
        """
        return self.model.get_interpreter()


class ActivityParameterTable(BaseParameterTable):
    MODEL = ActivityParameterModel

    def __init__(self, parent=None):
        super().__init__(parent)
        self.table_name = "activity_parameter"

        # Set delegates for specific columns
        self.setItemDelegateForColumn(1, delegates.FloatDelegate(self))
        self.setItemDelegateForColumn(2, delegates.FormulaDelegate(self))
        self.setItemDelegateForColumn(6, delegates.StringDelegate(self))
        self.setItemDelegateForColumn(7, delegates.ListDelegate(self))
        self.setItemDelegateForColumn(9, delegates.StringDelegate(self))
        self.setItemDelegateForColumn(10, delegates.ViewOnlyUncertaintyDelegate(self))

        # Set dropEnabled
        self.setDragDropMode(ABDataFrameView.DragDropMode.DropOnly)
        self.setAcceptDrops(True)

    def dragMoveEvent(self, event, /):
        pass

    def dragEnterEvent(self, event: QDragMoveEvent) -> None:
        """Check that the dragged row is from the databases table"""
        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            event.accept()

    def dropEvent(self, event: QDropEvent) -> None:
        """If the user drops an activity into the activity parameters table
        read the relevant data from the database and generate a new row.

        Also, create a warning if the activity is from a read-only database
        """
        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
        processes = set()

        for key in keys:
            act = bd.get_node(key=key)
            if isinstance(act, bf.Product):
                continue
            processes.add(key)
        event.accept()
        actions.ParameterNewAutomatic.run(processes)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        """Override and activate QTableView.contextMenuEvent()

        All possible menu events should be added and wired up here
        """
        if self.indexAt(event.pos()).row() == -1:
            return
        menu = QMenu(self)
        menu.addAction(icons.qicons.add, "Open activities", self.open_activity_tab)
        menu.addAction(self.rename_action)
        menu.addAction(self.delete_action)
        menu.addAction(self.modify_uncertainty_action)
        proxy = self.indexAt(event.pos())
        if proxy.isValid():
            param = self.get_parameter(proxy)
            if param.is_deletable():
                self.delete_action.setEnabled(True)
            else:
                self.delete_action.setEnabled(False)
            menu.exec_(event.globalPos())

    @Slot()
    def open_activity_tab(self):
        """Triggers the activity tab to open one or more activities."""
        for proxy in self.selectedIndexes():
            key = self.get_key(proxy)
            signals.safe_open_activity_tab.emit(key)

    def uncertainty_columns(self, show: bool):
        for i in range(10, 16):
            self.setColumnHidden(i, not show)

    def get_key(self, proxy=None) -> tuple:
        proxy = proxy or self.currentIndex()
        return self.model.get_key(proxy)

    def get_activity_groups(self, proxy, ignore_groups: list = None):
        return self.model.get_activity_groups(proxy, ignore_groups)

    @staticmethod
    def get_usable_parameters():
        return ActivityParameterModel.get_usable_parameters()

    def get_current_group(self, proxy=None) -> str:
        """Retrieve the group of the activity currently selected."""
        return self.model.get_group(proxy or self.currentIndex())

    def get_interpreter(self) -> Interpreter:
        return self.model.get_interpreter()


class ExchangesTable(ABDictTreeView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.model = ParameterTreeModel(parent=self)
        self.setModel(self.model)



from logging import getLogger

from qtpy import QtWidgets

from activity_browser import signals


log = getLogger(__name__)


class CentralTabWidget(QtWidgets.QTabWidget):
    """
    A custom QTabWidget that manages groups of tabs and their associated pages.

    This widget allows for organizing tabs into groups, dynamically adding pages to groups,
    and ensuring that each page has a unique object name.
    """

    def __init__(self, *args):
        """
        Initialize the CentralTabWidget.

        Args:
            *args: Positional arguments passed to the parent QTabWidget.
        """
        super().__init__(*args)
        # Connect to the project changed signal to reset the current index to 0
        signals.project.changed.connect(self.reset)

    @property
    def groups(self):
        """
        Retrieve all group widgets within the CentralTabWidget.

        Groups are identified as instances of GroupTabWidget.

        Returns:
            dict: A dictionary where keys are group object names and values are GroupTabWidget instances.
        """
        widgets = [self.widget(i) for i in range(self.count())]
        group_widgets = [widget for widget in widgets if isinstance(widget, GroupTabWidget)]
        return {widget.objectName(): widget for widget in group_widgets}

    def addToGroup(self, group: str, page: QtWidgets.QWidget):
        """
        Add a page to a specified group. If the group does not exist, it is created.

        Args:
            group (str): The name of the group to which the page will be added.
            page (QtWidgets.QWidget): The page to add to the group.

        Raises:
            ValueError: If the page does not have an object name.
        """
        if not page.objectName():
            raise ValueError("Page must have an object name")
        if group not in self.groups:
            # Create a new group if it does not exist
            self.addTab(GroupTabWidget(group, self), group)

        group = self.groups[group]

        # Check if the page already exists in the group
        page_names = [group.widget(i).objectName() for i in range(group.count())]
        if page.objectName() not in page_names:
            # Add the page to the group if it does not exist
            name = page.windowTitle() or page.objectName()  # Use windowTitle if available
            page.setWindowTitle(name)  # make sure the page has a title
            page.setParent(group)
            group.addTab(page, name)

            page.windowTitleChanged.connect(lambda title: group.setTabText(group.indexOf(page), title))
        else:
            # Set the existing page as the current tab
            index = page_names.index(page.objectName())
            group.setCurrentIndex(index)

        # Set the group and page as the current widgets
        self.setCurrentWidget(group)
        group.setCurrentWidget(page)

    def reset(self):
        self.setCurrentIndex(0)


class GroupTabWidget(QtWidgets.QTabWidget):
    """
    A custom QTabWidget that represents a group of tabs.

    This widget allows for managing tabs within a group, including making tabs movable,
    closable, and handling their lifecycle when the project changes or tabs are closed.
    """

    def __init__(self, name: str, *args):
        """
        Initialize the GroupTabWidget.

        Args:
            name (str): The name of the group, used as the object name for the widget.
            *args: Additional positional arguments passed to the parent QTabWidget.
        """
        super().__init__(*args)
        self.setMovable(True)  # Allow tabs to be rearranged.
        self.setTabsClosable(True)  # Allow tabs to be closed.
        self.setDocumentMode(True)  # Enable document mode for a more modern appearance.

        self.setObjectName(name)  # Set the object name for the widget.

        self.connect_signals()  # Connect necessary signals.

    def connect_signals(self):
        """
        Connect signals to their respective handlers.

        - Connects the `tabCloseRequested` signal to the `tabClosed` method.
        - Connects the `project.changed` signal to the `deleteLater` method to clean up the widget.
        """
        self.tabCloseRequested.connect(self.tabClosed)
        signals.project.changed.connect(self.deleteLater)

    def addTab(self, widget, *args, **kwargs):
        super().addTab(widget, *args, **kwargs)
        widget.destroyed.connect(self.checkEmpty)

    def checkEmpty(self):
        """
        Check if the GroupTabWidget is empty (i.e., has no tabs).

        If it is empty, delete the widget.
        """
        if self.count() == 1:
            self.deleteLater()

    def tabClosed(self, index):
        """
        Handle the closing of a tab.

        Deletes the widget associated with the tab and removes the tab from the widget.
        If no tabs remain, the entire GroupTabWidget is deleted.

        Args:
            index (int): The index of the tab to be closed.
        """
        self.widget(index).deleteLater()  # Delete the widget associated with the tab.




import os
from pathlib import Path
from typing import Iterable, Tuple

import requests
from qtpy import QtWidgets

from activity_browser.mod import bw2data as bd

from .settings import ab_settings


def get_base_path() -> Path:
    return Path(__file__).resolve().parents[0]


def read_file_text(file_dir: str) -> str:
    if not file_dir:
        raise ValueError("File path passed is empty")
    file = open(file_dir, mode="r", encoding="UTF-8")
    if not file:
        raise ValueError("File does not exist in the passed path:", file_dir)
    text = file.read()
    file.close()
    return text


def savefilepath(
    default_file_name: str = "AB_file", file_filter: str = "All Files (*.*)"
):
    """A central function to get a safe file path."""
    safe_name = bd.utils.safe_filename(default_file_name, add_hash=False)
    filepath, _ = QtWidgets.QFileDialog.getSaveFileName(
        parent=None,
        caption="Choose location for saving",
        dir=os.path.join(ab_settings.data_dir, safe_name),
        filter=file_filter,
    )
    return filepath


def safe_link_fetch(url: str) -> Tuple[object, object]:
    """
    Get a web-page or file from the internet or the error of getting the link.

    Parameters
    ----------
    url: a link

    Returns
    -------
    object: error if any, otherwise None
    object: response if no error, otherwise None
    """
    try:
        response = requests.get(url, timeout=2)  # retrieve the page from the URL
        response.raise_for_status()
    except Exception as error:
        return (None, error)

    return (response, None)


def sort_semantic_versions(versions: Iterable, highest_to_lowest: bool = True) -> list:
    """Return a sorted (default highest to lowest) list of semantic versions.

    Sorts based on the semantic versioning system.
    """
    return list(
        sorted(
            versions,
            key=lambda x: tuple(map(int, x.split("."))),
            reverse=highest_to_lowest,
        )
    )


def get_templates() -> dict:
    import platformdirs, os

    base_dir = platformdirs.user_data_dir(appname="ActivityBrowser", appauthor="ActivityBrowser")
    template_dir = os.path.join(base_dir, "templates")
    os.makedirs(template_dir, exist_ok=True)

    collection = {}

    for file in os.listdir(template_dir):
        if file.endswith(".tar.gz"):
            collection[file[:-7]] = os.path.join(template_dir, file)

    return collection




import logging
import os
import time
import sys
from traceback import extract_tb
from types import TracebackType
from typing import Type

import platformdirs


class ABFileHandler(logging.Handler):
    """
    LogHandler for the log files. Formats them in semicolon separated CSV files for easy reading.
    """

    headers = [
        "time",
        "type",
        "thread",
        "name",
        "file location",
        "line number",
        "function name",
        "message",
    ]

    def __init__(self):
        super().__init__()

        # create a unique filename based on the datetime
        self.filename = "ab_logs" + self.timestamp() + ".csv"

        # set dir and create it if it doesn't exist yet
        dir_path = str(platformdirs.user_log_dir(appname="ActivityBrowser", appauthor="ActivityBrowser"))
        os.makedirs(dir_path, exist_ok=True)

        # create final filepath of the logfile of this session
        self.filepath = os.path.join(dir_path, self.filename)

        # set the global file location
        global log_file_location
        log_file_location = self.filepath

        # create the logfile and write the headers
        with open(self.filepath, "a", encoding='utf-8') as log_file:
            log_file.write(";".join(self.headers) + "\n")

    def emit(self, record: logging.LogRecord):
        """Handle a new LogRecord"""
        # format the message from the record
        message = self.format(record)

        # append to the logfile
        with open(self.filepath, "a", encoding='utf-8') as log_file:
            log_file.write(message)

            # if there's exception info, write the exception traceback to the file as well
            if record.exc_info:
                exc_message = self.format_exception(record.exc_info[2])
                log_file.write(exc_message)

    def format(self, record: logging.LogRecord) -> str:
        """Format a LogRecord"""
        # format message to a single line
        message = " ".join(str(record.msg).split("\n"))
        message = message + " ".join([str(arg) for arg in record.args])

        # if there is no message left, return nothing
        if message == " ":
            return ""

        # make sure there a no semicolons
        message.replace(";", ":")

        # convert time
        struct_time = time.localtime(record.created)
        readable_time = time.strftime("%H:%M.%S", struct_time)

        line = f"{readable_time};{record.levelname};{record.threadName};{record.name};{record.pathname};{record.lineno};{record.funcName};{message}"
        return f"{line}\n"

    def format_exception(self, traceback: TracebackType) -> str:
        """Format the traceback of an exception"""
        # extract the traceback
        traceback = extract_tb(traceback)
        message = ""

        # append a line for each frame in the traceback
        for frame in traceback:
            line = f";TRACEBACK;;;{frame.filename};{frame.lineno};{frame.name};{frame.line}"
            message = f"{message}{line}\n"

        # return the string containing multiple lines
        return message

    def timestamp(self) -> str:
        """Return a timestamped string, the format provided is:
        day of the year _ month _ day - hour _ minute _ second"""
        stmp = time.localtime()
        return f"-{stmp.tm_year}-{stmp.tm_mon}-{stmp.tm_mday}_{stmp.tm_hour}-{stmp.tm_min}-{stmp.tm_sec}"


class ABPycharmHandler(logging.Handler):
    """
    LogHandler for the console. Make sure they are all in the same format. Adds badges, and if extended logs are enabled
    also the time and a (shortened) logger name.
    """

    badge = {
        "INFO": "\u001b[48;5;24m\u001b[38;5;255m INFO  \u001b[0m",
        "DEBUG": "\u001b[48;5;90m\u001b[38;5;255m DEBUG \u001b[0m",
        "EXCEPTION": "\u001b[48;5;88m\u001b[38;5;255m EXCPT \u001b[0m",
        "ERROR": "\u001b[48;5;88m\u001b[38;5;255m ERROR \u001b[0m",
        "WARNING": "\u001b[48;5;130m\u001b[38;5;255m WARN  \u001b[0m",
        "PRINT": "\u001b[7m PRINT \u001b[0m",
    }

    alias = {"activity_browser": "AB", "brightway2": "BW2"}

    def __init__(self):
        super().__init__()
        # create a unique filename based on the datetime
        self.filename = "pycharm_logs.log"

        # set dir and create it if it doesn't exist yet
        dir_path = platformdirs.user_log_dir("ActivityBrowser", "ActivityBrowser")
        os.makedirs(dir_path, exist_ok=True)

        # create final filepath of the logfile of this session
        self.filepath = os.path.join(dir_path, self.filename)

    def emit(self, record: logging.LogRecord):
        """Handle a new LogRecord"""
        # format message
        message = self.format_log(record)

        # append to the logfile
        with open(self.filepath, "a", encoding='utf-8') as log_file:
            log_file.write(message)

            # if there's exception info, write the exception traceback to the file as well
            if record.exc_info:
                exc_message = self.format_exception(record.exc_info[2])
                log_file.write(exc_message)

    def format_log(self, record: logging.LogRecord) -> str:
        """Format a LogRecord"""
        # format message to a single line
        message = " ".join(str(record.msg).split("\n"))
        message = message + " ".join([str(arg) for arg in record.args])

        # if there is no message left, return nothing
        if message == " ":
            return ""

        # clean-up if the message is a C++ error message
        if message.startswith("[") and message.index(":") < message.index("]"):
            # most likely a c++ error log, otherwise, very bad luck
            i = message.index("]") + 2
            message = message[i:]

        # retrieve the badge
        badge = self.badge[record.levelname]
        if record.exc_info:
            badge = self.badge["EXCEPTION"]

        # get a clean timestamp
        time_stamp = time.asctime()[11:19]

        source_str = self.format_source(record.name)

        return f"{time_stamp} {source_str}{badge} {message}\n"

    def format_source(self, name: str) -> str:
        """
        The entire source may be too long for the console window. Here we replace known sources with their alias, only
        use the first two modules and cut it short if it's still too long
        """
        # create list of the module string
        module_split = name.split(".")

        # switch a possible alias
        for key, alias in self.alias.items():
            if key == module_split[0]:
                module_split[0] = alias

        # rebuild only the first two modules
        source_str = ".".join(module_split[:2])

        # adjust length
        if len(source_str) >= 20:
            source_str = source_str[:16] + "..."

        return source_str.ljust(20)

    def format_exception(self, traceback: TracebackType) -> str:
        """Format the traceback of an exception"""
        space = 37

        traceback = extract_tb(traceback)
        message = "\u001b[38;5;1m\u001b[1m"
        for frame in traceback:
            line1 = f'{space * " "}File "{frame.filename}", line {frame.lineno}, in {frame.name}\n'
            line2 = f"{space * ' '} {frame.line}\n"
            message = message + line1 + line2
        message = f"{message}\u001b[0m"
        return message


def exception_hook(
    error: Type[BaseException], message: BaseException, traceback: TracebackType
):
    """Exception hook to catch and log exceptions"""
    exc_info = (error, message, traceback)
    log = logging.getLogger("exception_hook")
    log.exception(f"{error.__name__}: {message}", exc_info=exc_info)


def setup_ab_logging():
    # set the root logger's level to 0, this gives us access to all logs
    logging.root.setLevel(0)

    # peewee is mad, so set to info
    logging.getLogger("peewee").setLevel("INFO")

    # setting up a basic stderr handler
    stderr_handler = logging.StreamHandler()
    formatter = logging.Formatter(
        "%(asctime)s | %(levelname)s | %(message)s", "%H:%M:%S"
    )
    stderr_handler.setFormatter(formatter)
    stderr_handler.setLevel("DEBUG")
    logging.root.addHandler(stderr_handler)

    # # setting up the pycharm handler
    # pycharm_handler = ABPycharmHandler()
    # logging.root.addHandler(pycharm_handler)

    # setting up the file handler
    file_handler = ABFileHandler()
    logging.root.addHandler(file_handler)

    # setting up the exception hook
    sys.excepthook = exception_hook


log_file_location = None



import pandas as pd
from qtpy import QtCore, QtGui
from qtpy.QtCore import Qt, Signal, SignalInstance

from activity_browser.ui.icons import qicons

from .item import ABAbstractItem, ABBranchItem, ABDataItem


class ABItemModel(QtCore.QAbstractItemModel):
    grouped: SignalInstance = Signal(list)

    dataItemClass = ABDataItem
    branchItemClass = ABBranchItem

    def __init__(self, parent=None, dataframe=None):
        super().__init__(parent)

        if dataframe is None:
            dataframe = pd.DataFrame()

        self.dataframe: pd.DataFrame = dataframe  # DataFrame containing the visible data
        self.root: ABBranchItem = self.branchItemClass("root")  # root ABItem for the object tree
        self.grouped_columns: [int] = list()  # list of columns that are currently being grouped
        self.filtered_columns: [int] = set()  # set of all columns that have filters applied
        self.sort_column: int = -1  # column that is currently sorted
        self.sort_order: Qt.SortOrder = Qt.SortOrder.AscendingOrder
        self._query = ""  # Pandas query currently applied to the dataframe

        self.setDataFrame(self.dataframe)

    def columns(self):
        return [col for col in self.dataframe.columns if not str(col).startswith("_")]

    def headers(self):
        return self.columns()

    def index(self, row: int, column: int, parent: QtCore.QModelIndex = ...) -> QtCore.QModelIndex:
        """
        Create a QModelIndex based on a specific row, column and parent. Sets the associated ABItem as
        internalPointer. This will be the root ABItem if the parent is invalid.
        """
        # get the parent ABItem, or the root ABItem if the parent is invalid
        parent = parent.internalPointer() if parent.isValid() else self.root

        # get the child ABItem from the parent with the same rank as the specified row
        child = parent.iloc(row)

        # create and return a QModelIndex
        return self.createIndex(row, column, child)

    def indexFromPath(self, path: [str]) -> QtCore.QModelIndex:
        """
        Create a QModelIndex based on a specific path for the ABItem tree. The index column will be 0.
        """
        # get the ABItem for that specific path
        child = self.root.loc(path)
        if child is None:
            return QtCore.QModelIndex()

        # create and return a QModelIndex with the child's rank as row and 0 as column
        return self.createIndex(child.rank(), 0, child)

    def parent(self, child: QtCore.QModelIndex) -> QtCore.QModelIndex:
        """
        Return the parent of a QModelIndex.
        """
        if not child.isValid():
            return QtCore.QModelIndex()

        # get the ABItem from the QModelIndex
        child = child.internalPointer()

        # try to get the parent ABItem from the child
        try:
            parent = child.parent()
        # return an invalid/empty QModelIndex if this fails
        except:
            return QtCore.QModelIndex()

        # if the parent is the root ABItem return an invalid/empty QModelIndex
        if parent == self.root:
            return QtCore.QModelIndex()

        # create and return a QModelIndex with the child's rank as row and 0 as column
        return self.createIndex(parent.rank(), 0, parent)

    def rowCount(self, parent: QtCore.QModelIndex = ...) -> int:
        """
        Return the number of rows within the model
        """
        # return 0 if there is no DataFrame
        if self.dataframe is None:
            return 0
        # if the parent is the top of the table, the rowCount is the number of children for the root ABItem
        if not parent.isValid():
            value = len(self.root.children())
        # else it's the number of children within the ABItem saved within the internalPointer
        elif isinstance(parent.internalPointer(), ABAbstractItem):
            value = len(parent.internalPointer().children())
        # this shouldn't happen, but a failsafe
        else:
            value = 0
        return value

    def columnCount(self, parent: QtCore.QModelIndex = ...) -> int:
        """
        Return the number of columns within the model
        """
        # return 0 if there is no DataFrame
        if self.dataframe is None:
            return 0
        return len(self.columns())

    def data(self, index: QtCore.QModelIndex, role=Qt.ItemDataRole.DisplayRole):
        """
        Get the data associated with a specific index and role
        """
        if not index.isValid() or not isinstance(index.internalPointer(), ABAbstractItem):
            return None

        item: ABAbstractItem = index.internalPointer()
        col = index.column()
        key = self.columns()[col]

        # redirect to the item's displayData method
        if role == Qt.ItemDataRole.DisplayRole:
            return item.displayData(col, key)

        # redirect to the item's fontData method
        if role == Qt.ItemDataRole.FontRole:
            return item.fontData(col, key)

        # redirect to the item's decorationData method
        if role == Qt.ItemDataRole.DecorationRole:
            return item.decorationData(col, key)

        if role == Qt.ItemDataRole.BackgroundRole:
            return item.backgroundData(col, key)

        if role == Qt.ItemDataRole.ForegroundRole:
            return item.foregroundData(col, key)

        # else return None
        return None

    def setData(self, index: QtCore.QModelIndex, value, role=Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid() or not isinstance(index.internalPointer(), ABAbstractItem):
            return False

        if role == Qt.ItemDataRole.EditRole:
            success = index.internalPointer().setData(index.column(), self.columns()[index.column()], value)

            if success:
                self.dataChanged.emit(index, index, [Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole])
            return success

        return False

    def headerData(self, section, orientation=Qt.Orientation.Horizontal, role=Qt.ItemDataRole.DisplayRole):
        if orientation != Qt.Orientation.Horizontal:
            return None

        if role == Qt.ItemDataRole.DisplayRole:
            if section == 0 and self.grouped_columns:
                return " > ".join([self.headers()[column] for column in self.grouped_columns] + [self.headers()[0]])
            return self.headers()[section]

        if role == Qt.ItemDataRole.FontRole and section in self.filtered_columns:
            font = QtGui.QFont()
            font.setUnderline(True)
            return font

        if role == Qt.ItemDataRole.DecorationRole and section in self.filtered_columns:
            return qicons.filter

    def flags(self, index):
        if not index.isValid() or not isinstance(index.internalPointer(), ABAbstractItem):
            return Qt.ItemFlag.NoItemFlags
        if index.column() > len(self.columns()) - 1:
            return Qt.ItemFlag.NoItemFlags

        return index.internalPointer().flags(index.column(), self.columns()[index.column()])

    def endResetModel(self):
        """
        Reset the model based on dataframe, query and grouped columns. Should be called to reflect the changes of
        changing the dataframe, grouped columns or query string.
        """
        # if self.dataframe is None or self.dataframe.empty:
        #     return

        # apply any queries to the dataframe
        if q := self.query():
            df = self.dataframe.query(q).reset_index(drop=True).copy()
        else:
            df = self.dataframe.copy()

        if not self.sort_column > len(self.columns()) - 1 and self.sort_column != -1:
            # apply the sorting
            df.sort_values(
                by=self.columns()[self.sort_column],
                ascending=(self.sort_order == Qt.SortOrder.AscendingOrder),
                inplace=True, ignore_index=True
            )

        # rebuild the ABItem tree
        self.root = self.branchItemClass("root")
        items = self.createItems(df)

        # if no grouping of Entries, just append everything as a direct child of the root ABItem
        if not self.grouped_columns:
            for i, item in enumerate(items):
                item.set_parent(self.root)
        # else build paths based on the grouped columns and create an ABItem tree
        else:
            column_names = [self.columns()[column] for column in self.grouped_columns]

            for i, *paths in df[column_names].itertuples():
                joined_path = []

                for path in paths:
                    joined_path.extend(path) if isinstance(path, (list, tuple)) else joined_path.append(path)

                joined_path.append(i)
                self.root.put(items[df.index.get_loc(i)], joined_path)

        super().endResetModel()

    def createItems(self, dataframe=None) -> list["ABAbstractItem"]:
        if dataframe is None:
            dataframe = self.dataframe
        return [self.dataItemClass(index, data) for index, data in dataframe.to_dict(orient="index").items()]

    def setDataFrame(self, dataframe: pd.DataFrame):
        self.beginResetModel()
        self.dataframe = dataframe
        self.endResetModel()

    def sort(self, column: int, order=Qt.SortOrder.AscendingOrder):
        if column + 1 > len(self.columns()):
            return
        if column == self.sort_column and order == self.sort_order:
            return

        self.beginResetModel()

        self.sort_column = column
        self.sort_order = order

        self.endResetModel()

    def group(self, column: int):
        self.beginResetModel()
        self.grouped_columns.append(column)
        self.endResetModel()
        self.grouped.emit(self.grouped_columns)

    def ungroup(self):
        self.beginResetModel()
        self.grouped_columns.clear()
        self.endResetModel()
        self.grouped.emit(self.grouped_columns)

    def query(self) -> str:
        return self._query

    def setQuery(self, query: str):
        """Apply the query string to the dataframe and rebuild the model"""
        self.beginResetModel()
        self._query = query
        self.endResetModel()

    def hasChildren(self, parent: QtCore.QModelIndex):
        item = parent.internalPointer()
        if isinstance(item, ABAbstractItem):
            return item.has_children()
        return super().hasChildren(parent)






# -*- coding: utf-8 -*
"""
Code in `pedigree.py` is taken wholesale from the https://bitbucket.org/cmutel/pedigree-matrix/src repository,
which is published by Chris Mutel under an MIT license (2012).

FUTURE WORK?
'The application of the pedigree approach to the distributions foreseen in ecoinvent v3'
(doi: 10.1007/s11367-014-0759-5) reveals the formulas used to convert the calculated sigma
(+ geometric standard deviation) into the coefficient of variation ('CV').
In turn, this CV can then be used to generate required values for a number of other uncertainty distributions.

Any additions made should improve the transition of the calculated sigma (or Geometric Standard Deviation/GSD)
smoothly into the related uncertainty distributions.
"""
import math
from pprint import pformat

from bw2data.parameters import ParameterBase
from bw2data.proxies import ExchangeProxyBase

VERSION_2 = {
    "reliability": (1.0, 1.54, 1.61, 1.69, 1.69),
    "completeness": (1.0, 1.03, 1.04, 1.08, 1.08),
    "temporal correlation": (1.0, 1.03, 1.1, 1.19, 1.29),
    "geographical correlation": (1.0, 1.04, 1.08, 1.11, 1.11),
    "further technological correlation": (1.0, 1.18, 1.65, 2.08, 2.8),
    "sample size": (1.0, 1.0, 1.0, 1.0, 1.0),
}


class PedigreeMatrix(object):
    __slots__ = ["factors"]
    labels = (
        "reliability",
        "completeness",
        "temporal correlation",
        "geographical correlation",
        "further technological correlation",
        "sample size",
    )

    def __init__(self):
        self.factors = {}

    @classmethod
    def from_numbers(cls, data: tuple) -> "PedigreeMatrix":
        """Takes a tuple of integers and construct a PedigreeMatrix."""
        assert len(data) in (5, 6), "Must provide either 5 or 6 factors"
        if len(data) == 5:
            data = data + (1,)
        matrix = cls()
        for index, factor in enumerate(data):
            matrix.factors[cls.labels[index]] = factor
        return matrix

    @classmethod
    def from_dict(cls, data: dict) -> "PedigreeMatrix":
        return cls.from_numbers(tuple(data.get(k) for k in cls.labels if k in data))

    @classmethod
    def from_bw_object(cls, obj) -> "PedigreeMatrix":
        if isinstance(obj, ExchangeProxyBase):
            return cls.from_dict(obj.get("pedigree", {}))
        elif isinstance(obj, ParameterBase) and "pedigree" in obj.data:
            return cls.from_dict(obj.data.get("pedigree", {}))
        else:
            raise AssertionError("Could not find pedigree in object")

    def calculate(
        self, basic_uncertainty: float = 1.0, as_geometric_sigma: bool = False
    ) -> float:
        """Calculates the sigma or geometric standard deviation from the factors."""
        values = [basic_uncertainty] + self.get_values()
        sigma = math.sqrt(sum([math.log(x) ** 2 for x in values])) / 2
        return sigma if not as_geometric_sigma else math.exp(2 * sigma)

    def get_values(self) -> list:
        assert self.factors, "Must provide Pedigree Matrix factors"
        return [VERSION_2[key][index - 1] for key, index in self.factors.items()]

    def factors_as_tuple(self):
        return tuple(self.factors[k] for k in self.labels if k in self.factors)

    def __repr__(self) -> str:
        return "Empty Pedigree Matrix" if not self.factors else pformat(self.factors)



from .activity.activity_relink import ActivityRelink
from .activity.activity_duplicate import ActivityDuplicate
from .activity.activity_open import ActivityOpen
from .activity.activity_graph import ActivityGraph
from .activity.activity_duplicate_to_loc import ActivityDuplicateToLoc
from .activity.activity_delete import ActivityDelete
from .activity.activity_duplicate_to_db import ActivityDuplicateToDB
from .activity.activity_modify import ActivityModify
from .activity.activity_new_process import ActivityNewProcess
from .activity.activity_new_product import ActivityNewProduct
from .activity.activity_open import ActivityOpen
from .activity.activity_relink import ActivityRelink
from .activity.activity_sdf_to_clipboard import ActivitySDFToClipboard
from .activity.activity_redo_allocation import MultifunctionalProcessRedoAllocation
from .activity.process_property_modify import ProcessPropertyModify
from .activity.process_property_remove import ProcessPropertyRemove

from .calculation_setup.cs_new import CSNew
from .calculation_setup.cs_delete import CSDelete
from .calculation_setup.cs_duplicate import CSDuplicate
from .calculation_setup.cs_rename import CSRename
from .calculation_setup.cs_add_functional_unit import CSAddFunctionalUnit
from .calculation_setup.cs_change_functional_unit import CSChangeFunctionalUnit
from .calculation_setup.cs_add_impact_category import CSAddImpactCategory
from .calculation_setup.cs_delete_impact_category import CSDeleteImpactCategory
from .calculation_setup.cs_delete_functional_unit import CSDeleteFunctionalUnit
from .calculation_setup.cs_calculate import CSCalculate
from .calculation_setup.cs_open import CSOpen

from .database.database_open import DatabaseOpen
from .database.database_export import DatabaseExport
from .database.database_new import DatabaseNew
from .database.database_delete import DatabaseDelete
from .database.database_duplicate import DatabaseDuplicate
from .database.database_relink import DatabaseRelink
from .database.database_redo_allocation import DatabaseRedoAllocation
from .database.database_explorer_open import DatabaseExplorerOpen
from .database.database_process import DatabaseProcess
from .database.database_import_from_ecoinvent import DatabaseImportFromEcoinvent
from .database.database_importer_excel import DatabaseImporterExcel
from .database.database_importer_bw2package import DatabaseImporterBW2Package
from .database.database_set_readonly import DatabaseSetReadonly

from .exchange.exchange_copy_sdf import ExchangeCopySDF

from .exchange.exchange_new import ExchangeNew
from .exchange.exchange_delete import ExchangeDelete
from .exchange.exchange_modify import ExchangeModify
from .exchange.exchange_formula_remove import ExchangeFormulaRemove
from .exchange.exchange_uncertainty_modify import ExchangeUncertaintyModify
from .exchange.exchange_uncertainty_remove import ExchangeUncertaintyRemove
from .exchange.exchange_copy_sdf import ExchangeCopySDF
from .exchange.exchange_sdf_to_clipboard import ExchangeSDFToClipboard

from .method.method_duplicate import MethodDuplicate
from .method.method_delete import MethodDelete
from .method.method_open import MethodOpen
from .method.method_rename import MethodRename
from .method.method_meta_modify import MethodMetaModify

from .method.importer.method_importer_ecoinvent import MethodImporterEcoinvent
from .method.importer.method_importer_bw2io import MethodImporterBW2IO

from .method.cf_uncertainty_modify import CFUncertaintyModify
from .method.cf_amount_modify import CFAmountModify
from .method.cf_remove import CFRemove
from .method.cf_new import CFNew
from .method.cf_uncertainty_remove import CFUncertaintyRemove

from .parameter.parameter_new import ParameterNew
from .parameter.parameter_new_automatic import ParameterNewAutomatic
from .parameter.parameter_new_from_parameter import ParameterNewFromParameter
from .parameter.parameter_rename import ParameterRename
from .parameter.parameter_delete import ParameterDelete
from .parameter.parameter_modify import ParameterModify
from .parameter.parameter_uncertainty_remove import ParameterUncertaintyRemove
from .parameter.parameter_uncertainty_modify import ParameterUncertaintyModify
from .parameter.parameter_clear_broken import ParameterClearBroken

from .project.project_new import ProjectNew
from .project.project_duplicate import ProjectDuplicate
from .project.project_delete import ProjectDelete
from .project.project_duplicate import ProjectDuplicate
from .project.project_remote_import import ProjectRemoteImport
from .project.project_local_import import ProjectLocalImport
from .project.project_new import ProjectNew
from .project.project_new_remote import ProjectNewRemote
from .project.project_switch import ProjectSwitch
from .project.project_export import ProjectExport
from .project.project_import import ProjectImport
from .project.project_manager_open import ProjectManagerOpen
from .project.project_migrate25 import ProjectMigrate25
from .project.project_create_template import ProjectCreateTemplate
from .project.project_new_template import ProjectNewFromTemplate

from .settings_wizard_open import SettingsWizardOpen
from .migrations_install import MigrationsInstall
from .pyside_upgrade import PysideUpgrade
from .metadatastore_open import MetaDataStoreOpen



from qtpy import QtWidgets

from bw2data.parameters import ParameterBase, parameters

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from activity_browser.bwutils import Parameter, refresh_parameter


class ParameterRename(ABAction):
    """
    ABAction to rename an existing parameter. Constructs a dialog for the user in which they choose the new name. If no
    name is chosen, or the user cancels: return. Else, instruct the ParameterController to rename the parameter using
    the given name.
    """

    icon = qicons.edit
    text = "Rename parameter..."

    @staticmethod
    @exception_dialogs
    def run(parameter: tuple | Parameter | ParameterBase, new_name: str = None):
        parameter = refresh_parameter(parameter)

        new_name = new_name or ParameterRename.get_new_name(parameter)

        if not new_name:
            return

        if not new_name.isidentifier():
            raise ValueError("Parameter name must be a valid Python identifier")

        getattr(parameters, f"rename_{parameter.param_type}_parameter")(
                parameter.to_peewee_model(), new_name, update_dependencies=True
            )

    @staticmethod
    def get_new_name(parameter: Parameter):
        new_name, ok = QtWidgets.QInputDialog.getText(
            application.main_window,
            "Rename parameter",
            f"Rename parameter '{parameter.name}' to:",
        )

        if ok and new_name:
            return new_name



# -*- coding: utf-8 -*-
import itertools
import sqlite3
import pickle
from time import time
from functools import lru_cache
from typing import Set
from logging import getLogger

from playhouse.shortcuts import model_to_dict

import pandas as pd

from qtpy.QtCore import Qt, QObject, Signal, SignalInstance

import bw2data as bd
from bw2data.errors import UnknownObject
from bw2data.backends import sqlite3_lci_db, ActivityDataset

from activity_browser import signals


log = getLogger(__name__)


def list_to_tuple(x) -> tuple:
    return tuple(x) if isinstance(x, list) else x


class MetaDataStore(QObject):
    """A container for technosphere and biosphere metadata during an AB session.

    This is to prevent multiple time-expensive repetitions such as the code
    below at various places throughout the AB:

    .. code-block:: python
        meta_data = list()  # or whatever container
        for ds in bw.Database(name):
            meta_data.append([ds[field] for field in fields])

    Instead, this data store features a dataframe that contains all metadata
    and can be indexed by (activity or biosphere key).
    The columns feature the metadata.

    Properties
    ----------
    index

    """
    # Options for reading classification systems from ecoinvent databases are
    # - ISIC rev.4 ecoinvent
    # - CPC
    # - EcoSpold01Categories
    # - HS (>= ecoinvent 3.10)
    # To show these columns in `ActivitiesBiosphereModel`,
    # add them to `self.act_fields` there and `CLASSIFICATION_SYSTEMS` below
    CLASSIFICATION_SYSTEMS = ["ISIC rev.4 ecoinvent"]

    synced: SignalInstance = Signal()

    def __init__(self, parent=None):
        from activity_browser import application
        super().__init__(parent)
        self.dataframe = pd.DataFrame()
        self.moveToThread(application.thread())
        self.connect_signals()

    def connect_signals(self):
        signals.project.changed.connect(self.sync)
        signals.node.changed.connect(self.on_node_changed)
        signals.node.deleted.connect(self.on_node_deleted)
        signals.meta.databases_changed.connect(self.sync_databases)
        signals.database.deleted.connect(self.sync_databases)

    def on_node_deleted(self, ds):
        try:
            self.dataframe.drop(ds.key, inplace=True)
            self.synced.emit()
        except KeyError:
            pass

    def on_node_changed(self, new, old):
        data_raw = model_to_dict(new)
        data = data_raw.pop("data")
        data.update(data_raw)
        data["key"] = new.key
        data = pd.DataFrame([data], index=pd.MultiIndex.from_tuples([new.key]))

        if new.key in self.dataframe.index:  # the activity has been modified

            compare_old = self.dataframe.loc[new.key].dropna().sort_index()
            compare_new = data.loc[new.key].dropna().sort_index()

            if list(compare_new.index) == list(compare_old.index) and (compare_new == compare_old).all():
                return  # but it is the same as the current DF, so no sync necessary
            for col in [col for col in data.columns if col not in self.dataframe.columns]:
                self.dataframe[col] = pd.NA
            self.dataframe.loc[new.key] = data.loc[new.key]
        elif self.dataframe.empty:  # an activity has been added and the dataframe was empty
            self.dataframe = data
        else:  # an activity has been added and needs to be concatenated to existing metadata
            self.dataframe = pd.concat([self.dataframe, data], join="outer")

        self.thread().eventDispatcher().awake.connect(self._emitSyncLater, Qt.ConnectionType.UniqueConnection)

    @property
    def databases(self):
        return set(self.dataframe.get("database", []))

    def sync_node(self, key: tuple) -> None:
        """Update metadata when an activity has changed.

        Three situations:
        1. An activity has been deleted.
        2. Activity data has been modified.
        3. An activity has been added.
           Note that duplicating activities is the same as adding a new activity.

        Parameters
        ----------
        key : tuple
            The specific activity to update in the MetaDataStore
        """
        data = self._get_node(key)
        if data is None:  # the activity has been deleted
            self.dataframe.drop(key, inplace=True)
        elif key in self.dataframe.index:  # the activity has been modified
            self.dataframe.loc[key] = data.loc[key]
        else:  # an activity has been added
            self.dataframe = pd.concat([self.dataframe, data], join="outer")

        self.thread().eventDispatcher().awake.connect(self._emitSyncLater, Qt.ConnectionType.UniqueConnection)

    def _emitSyncLater(self):
        t = time()
        self.synced.emit()
        log.debug(f"Metadatastore sync signal completed in {time() - t:.2f} seconds")
        self.thread().eventDispatcher().awake.disconnect(self._emitSyncLater)

    def _get_node(self, key: tuple):
        try:
            id = bd.mapping[key]
        except (UnknownObject, ActivityDataset.DoesNotExist):
            return None

        con = sqlite3.connect(sqlite3_lci_db._filepath)
        node_df = pd.read_sql(f"SELECT * FROM activitydataset WHERE id = {id}", con)
        con.close()

        return self._parse_df(node_df)

    def sync_databases(self) -> None:
        sync = False

        for db_name in [x for x in self.databases if x not in bd.databases]:
            # deleted databases
            self.dataframe.drop(db_name, level=0, inplace=True)
            sync = True

        for db_name in [x for x in bd.databases if x not in self.databases]:
            # new databases
            data = self._get_database(db_name)
            if data is None:
                continue

            if self.dataframe.empty:
                self.dataframe = data
            else:
                self.dataframe = pd.concat([self.dataframe, data], join="outer")

            sync = True

        if sync:
            self.thread().eventDispatcher().awake.connect(self._emitSyncLater, Qt.ConnectionType.UniqueConnection)

    def _get_database(self, db_name: str) -> pd.DataFrame | None:
        con = sqlite3.connect(sqlite3_lci_db._filepath)
        node_df = pd.read_sql(f"SELECT * FROM activitydataset WHERE database = '{db_name}'", con)
        con.close()
        if node_df.empty:
            return None
        return self._parse_df(node_df)

    def sync(self) -> None:
        """Deletes metadata when the project is changed."""
        log.debug("Synchronizing MetaDataStore")

        con = sqlite3.connect(sqlite3_lci_db._filepath)
        node_df = pd.read_sql("SELECT * FROM activitydataset", con)
        con.close()

        self.dataframe = self._parse_df(node_df)

        self.synced.emit()

    def _parse_df(self, raw_df: pd.DataFrame) -> pd.DataFrame:
        data_df = pd.DataFrame([pickle.loads(x) for x in raw_df["data"]]).drop(columns=["id"], errors="ignore")

        df = raw_df.combine_first(data_df)
        df.drop(columns=["data"], inplace=True)

        df["key"] = df.loc[:, ["database", "code"]].apply(tuple, axis=1)
        if df.empty:
            return df
        df.index = pd.MultiIndex.from_tuples(df["key"])
        return df


    def get_existing_fields(self, field_list: list) -> list:
        """Return a list of fieldnames that exist in the current dataframe."""
        return [fn for fn in field_list if fn in self.dataframe.columns]

    def get_metadata(self, keys: list, columns: list) -> pd.DataFrame:
        """Return a slice of the dataframe matching row and column identifiers.

        NOTE: https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#deprecate-loc-reindex-listlike
        From pandas version 1.0 and onwards, attempting to select a column
        with all NaN values will fail with a KeyError.
        """
        df = self.dataframe.loc[pd.IndexSlice[keys], :]
        return df.reindex(columns, axis="columns")

    def get_metadata_from_ids(self, ids: list, columns: list) -> pd.DataFrame:
        """Return a slice of the dataframe matching row and column identifiers.

        NOTE: https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#deprecate-loc-reindex-listlike
        From pandas version 1.0 and onwards, attempting to select a column
        with all NaN values will fail with a KeyError.
        """
        df = self.dataframe.loc[self.dataframe["id"].isin(ids), :]
        return df.reindex(columns, axis="columns")

    def get_database_metadata(self, db_name: str) -> pd.DataFrame:
        """Return a slice of the dataframe matching the database.

        If the database does not exist in the metadata, attempt to add it.

        Parameters
        ----------
        db_name : str
            Name of the database to be retrieved

        Returns
        -------
        pd.DataFrame
            Slice of the metadata matching the database name

        """
        if db_name not in self.databases:
            return pd.DataFrame(columns=["name", "type", "location", "database", "code", "key", ])
        return self.dataframe.loc[self.dataframe["database"] == db_name].copy(deep=True).dropna(how='all', axis=1)

    @property
    def index(self):
        """Returns the (multi-) index of the MetaDataStore.

        This allows us to 'hide' the dataframe object in de AB_metadata
        """
        return self.dataframe.index

    def get_locations(self, db_name: str) -> set:
        """Returns a set of locations for the given database name."""
        data = self.get_database_metadata(db_name)
        if "location" not in data.columns:
            return set()
        locations = data["location"].unique()
        return set(locations[locations != ""])

    def get_units(self, db_name: str) -> set:
        """Returns a set of units for the given database name."""
        data = self.get_database_metadata(db_name)
        if "unit" not in data.columns:
            return set()
        units = data["unit"].unique()
        return set(units[units != ""])

    @lru_cache(maxsize=10)
    def get_tag_names_for_db(self, db_name: str) -> Set[str]:
        """Returns a set of tag names for the given database name."""
        data = self.get_database_metadata(db_name)
        if "tags" not in data.columns:
            return set()
        tags = data.tags.drop_duplicates().values
        tag_names = set(
            itertools.chain(*map(lambda x: x.keys(), filter(lambda x: x, tags)))
        )
        return tag_names

    def get_tag_names(self):
        """Returns a set of tag names for all databases."""
        tag_names = set()
        for db_name in self.databases:
            tag_names = tag_names.union(self.get_tag_names_for_db(db_name))
        return tag_names

    def print_convenience_information(self, db_name: str) -> None:
        """Reports how many unique locations and units the database has."""
        log.debug(
            "{} unique locations and {} unique units in {}".format(
                len(self.get_locations(db_name)), len(self.get_units(db_name)), db_name
            )
        )

    def unpack_classifications(self, df: pd.DataFrame, systems: list) -> pd.DataFrame:
        """Unpack classifications column to a new column for every classification system in 'systems'.

        Will return dataframe with added column.
        """
        classifications = list(df['classifications'].values)
        system_cols = []
        for system in systems:
            system_cols.append(self._unpacker(classifications, system))
        # creating the DF rotated is easier so we do that and then transpose
        unpacked = pd.DataFrame(system_cols, columns=df.index, index=systems).T

        # Finally, merge the df with the new unpacked df using indexes
        df = pd.merge(
            df, unpacked, how='inner', left_index=True,
            right_index=True, sort=False
        )
        return df

    def _unpacker(self, classifications: list, system: str) -> list:
        """Iterate over all 'c' lists in 'classifications'
        and add those matching 'system' to list 'system_classifications', when no matches, add empty string.
        If 'c' is not a list, add empty string.

        Always returns a list 'system_classifications' where len(system_classifications) == len(classifications).

        Testing showed that converting to list and doing the checks on a list is ~5x faster than keeping
        data in DF and using a df.apply() function, we do this now (difference was ~0.4s vs ~2s).
        """
        system_classifications = []
        for c in classifications:
            result = ""
            if not isinstance(c, (list, tuple, set)):
                system_classifications.append(result)
                continue
            for s in c:
                if s[0] == system:
                    result = s[1]
                    break
            system_classifications.append(result)  # result is either "" or the classification
        return system_classifications


AB_metadata = MetaDataStore()
